from copy import deepcopy

import numpy as np

QUANTITIES = {
    "zmass": {
        "label": r"${m_{\mathrm{Z}}}$ / GeV",
        "scale": "linear",
        "range": (70, 110),
        "gen_name": "genzmass",
        "gen_label": r"${m_{\mathrm{Z}}^{\mathrm{gen}}}$ / GeV",
    },
    "zphi": {
        "label": r"${\phi^{\mathrm{Z}}}$",
        "scale": "linear",
        "range": (-np.pi, np.pi),
        "gen_name": "genzphi",
        "gen_label": r"${\phi^{\mathrm{Z,gen}}}$",
    },
    "zpt": {
        "label": r"${p_{\mathrm{T}}^{\mathrm{Z}}}$ /  GeV",
        "range": (25, 1000),
        "scale": "log",
        "gen_name": "genzpt",
        "gen_label": r"${p_{\mathrm{T}}^{\mathrm{Z,gen}}}$ /  GeV",
        "xs_label": r"$\frac{{\mathrm{{d}}^3\sigma}}{{\mathrm{{d}}{{{}}}\,\mathrm{{d}}y_{{\mathrm{{b}}}}\,\mathrm{{d}}y*}}\,\,/\,\,\mathrm{{pb}}\,\,/\,\,\mathrm{{GeV}}\,\,\times\,X$".format(
            r"{p_{\mathrm{T}}^{\mathrm{Z}}}"
        ),
        "global_index_name": "binIndexzpt",
        "gen_global_index_name": "binIndexgenzpt",
    },
    "zy": {
        "label": r"$y^{\mathrm{Z}}$ Bin",
        "scale": "linear",
        "range": (-2.5, 2.5),
        "gen_label": r"$\mathrm{y}^{\mathrm{Z,gen}}}$ Bin",
        "gen_name": "genzy",
    },
    "binIndexzpt": {
        "label": r"Global bin ($y^{{*}}$, $y_{{\mathrm{{b}}}}$, $p_{\mathrm{T}}^{\mathrm{Z}}$)",
        "range": (0, 265),
        "gen_name": "binIndexgenzpt",
        "reco_label": r"Global reco. bin number ($y^{{*}}$, $y_{{\mathrm{{b}}}}$, $p_{\mathrm{T}}^{\mathrm{Z}}$)",
        "gen_label": r"Global gen. bin number ($y^{{*, \mathrm{{gen}}}}$, $y_{{\mathrm{{b}}}}^{{\mathrm{{gen}}}}$, $p_{\mathrm{T}}^{\mathrm{Z, gen}}$)",
    },
    "jet1eta": {
        "label": r"${\eta^{\mathrm{jet1}}}$",
        "scale": "linear",
        "range": (-2.4, 2.4),
        "gen_name": "genjet1eta",
        "gen_label": r"$\eta^{\mathrm{jet1,gen}}$",
    },
    "jet1phi": {
        "label": r"${\phi^{\mathrm{jet1}}}$",
        "scale": "linear",
        "range": (-np.pi, np.pi),
        "gen_name": "genjet1phi",
        "gen_label": r"${\phi^{\mathrm{jet1,gen}}}$",
    },
    "jet1y": {
        "label": r"$y^{\mathrm{jet1}}$ Bin",
        "scale": "linear",
        "range": (-2.4, 2.4),
        "gen_label": r"$\mathrm{y}^{\mathrm{jet1,gen}}}$ Bin",
        "gen_name": "genjet1y",
    },
    "jet1pt": {
        "label": r"${p_{\mathrm{T}}^{\mathrm{jet1}}}$ /  GeV",
        "scale": "log",
        "range": (20, 1500),
        "gen_name": "genjet1pt",
        "gen_label": r"${p_{\mathrm{T}}^{\mathrm{jet1,gen}}}$ /  GeV",
    },
    "muplusphi": {
        "label": r"${\phi^{\mu^{+}}}$",
        "scale": "linear",
        "range": (-np.pi, np.pi),
        "gen_name": "genmuplusphi",
        "gen_label": r"${\phi^{\mu^{+,gen}}}$",
    },
    "muminusphi": {
        "label": r"${\phi^{\mu^{-}}}$",
        "scale": "linear",
        "range": (-np.pi, np.pi),
        "gen_name": "genmuminusphi",
        "gen_label": r"${\phi^{\mu^{-,gen}}}$",
    },
    "mupluseta": {
        "label": r"${\eta^{\mu^{+}}}$",
        "scale": "linear",
        "range": (-2.4, 2.4),
        "gen_name": "genmupluseta",
        "gen_label": r"${\eta^{\mu^{+,gen}}}$",
    },
    "muminuseta": {
        "label": r"${\eta^{\mu^{-}}}$",
        "scale": "linear",
        "range": (-2.4, 2.4),
        "gen_name": "genmuminuseta",
        "gen_label": r"${\eta^{\mu^{-,gen}}}$",
    },
    "mupluspt": {
        "label": r"${p_{\mathrm{T}}^{\mu^{+}}}$",
        "scale": "linear",
        "range": (25, 500),
        "gen_name": "genmupluspt",
        "gen_label": r"${p_{\mathrm{T}}^{\mu^{+,gen}}}$",
    },
    "muminuspt": {
        "label": r"${p_{\mathrm{T}}^{\mu^{-}}}$",
        "scale": "linear",
        "range": (25, 500),
        "gen_name": "genmuminuspt",
        "gen_label": r"${p_{\mathrm{T}}^{\mu^{-}}}$",
    },
    "mu1phi": {
        "label": r"${\phi^{\mu_{1}}}$",
        "scale": "linear",
        "range": (-np.pi, np.pi),
        "gen_name": "genmu1phi",
        "gen_label": r"${\phi^{\mu_{1,gen}}}$",
    },
    "mu2phi": {
        "label": r"${\phi^{\mu_{2}}}$",
        "scale": "linear",
        "range": (-np.pi, np.pi),
        "gen_name": "genmu2phi",
        "gen_label": r"${\phi^{\mu_{2,gen}}}$",
    },
    "mu1eta": {
        "label": r"${\eta^{\mu_{1}}}$",
        "scale": "linear",
        "range": (-2.4, 2.4),
        "gen_name": "genmu1eta",
        "gen_label": r"${\eta^{\mu_{1,gen}}}$",
    },
    "mu2eta": {
        "label": r"${\eta^{\mu_{2}}}$",
        "scale": "linear",
        "range": (-2.4, 2.4),
        "gen_name": "genmu2eta",
        "gen_label": r"${\eta^{\mu_{2,gen}}}$",
    },
    "mu1pt": {
        "label": r"${p_{\mathrm{T}}^{\mu_{1}}}$",
        "scale": "linear",
        "range": (25, 500),
        "gen_name": "genmu1pt",
        "gen_label": r"${p_{\mathrm{T}}^{\mu_{1,gen}}}$",
    },
    "mu2pt": {
        "label": r"${p_{\mathrm{T}}^{\mu_{2}}}$",
        "scale": "linear",
        "range": (25, 500),
        "gen_name": "genmu2pt",
        "gen_label": r"${p_{\mathrm{T}}^{\mu_{2,gen}}}$",
    },
}

lumis = {
    "2016preVFP": 19.50,
    "2016preVFP_B": 5.83,
    "2016preVFP_C": 2.60,
    "2016preVFP_D": 4.29,
    "2016preVFP_E": 4.07,
    "2016preVFP_F": 2.72,
    "2016postVFP": 16.81,
    "2016postVFP_F": 0.42,
    "2016postVFP_G": 7.65,
    "2016postVFP_H": 8.74,
    "2016": 36.31,
    "2017": 41.48,
    "2017_B": 4.80,
    "2017_C": 9.57,
    "2017_D": 4.25,
    "2017_E": 9.31,
    "2017_F": 13.54,
    "2018": 59.83,
    "2018_A": 14.03,
    "2018_B": 7.07,
    "2018_C": 6.90,
    "2018_D": 31.84,
    "Run2": 137.64,
}

central_binning = np.array(
    [
        25.0,
        30.0,
        35.0,
        40.0,
        45.0,
        50.0,
        60.0,
        70.0,
        80.0,
        90.0,
        100.0,
        110.0,
        130.0,
        150.0,
        170.0,
        190.0,
        220.0,
        250.0,
        400.0,
        1000.0,
    ]
)

edge_binning = np.array(
    [
        25.0,
        30.0,
        35.0,
        40.0,
        45.0,
        50.0,
        60.0,
        70.0,
        80.0,
        90.0,
        100.0,
        110.0,
        130.0,
        150.0,
        170.0,
        190.0,
        250.0,
        1000.0,
    ]
)

extreme_binning = np.array([25.0, 30.0, 40.0, 50.0, 70.0, 90.0, 110.0, 150.0, 250.0])

ybys_bins = {
    "YB_00_05_YS_00_05": {
        "ys_range": (0.0, 0.5),
        "yb_range": (0.0, 0.5),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 12,
        "shift": 3.2,
        "color": "#5c15b6",
        "marker": "D",
        "marker_style": "full",
        "label": "$0.0 \\leq y_{\\mathrm{b}} < 0.5$\n$0.0 \\leq y{*} < 0.5$",
        "zpt_binning": central_binning,
    },
    "YB_05_10_YS_00_05": {
        "ys_range": (0.0, 0.5),
        "yb_range": (0.5, 1.0),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 11,
        "shift": 3.0,
        "color": "#7959c4",
        "marker": "D",
        "marker_style": "left",
        "label": "$0.5 \\leq y_{\\mathrm{b}} < 1.0$\n$0.0 \\leq y{*} < 0.5$",
        "zpt_binning": central_binning,
    },
    "YB_10_15_YS_00_05": {
        "ys_range": (0.0, 0.5),
        "yb_range": (1.0, 1.5),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 10,
        "shift": 2.8,
        "color": "#9983cf",
        "marker": "D",
        "marker_style": "right",
        "zpt_binning": central_binning,
        "label": "$1.0 \\leq y_{\\mathrm{b}} < 1.5$\n$0.0 \\leq y{*} < 0.5$",
    },
    "YB_15_20_YS_00_05": {
        "ys_range": (0.0, 0.5),
        "yb_range": (1.5, 2.0),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 9,
        "shift": 2.6,
        "color": "#c6b2e2",
        "marker": "D",
        "marker_style": "bottom",
        "label": "$1.5 \\leq y_{\\mathrm{b}} < 2.0$\n$0.0 \\leq y{*} < 0.5$",
        "zpt_binning": central_binning,
    },
    "YB_20_25_YS_00_05": {
        "ys_range": (0.0, 0.5),
        "yb_range": (2.0, 2.5),
        "pt_range": (25, 250),
        "area": 0.25,
        "stagger": 9,
        "shift": 2.4,
        "color": "#fab0ff",
        "marker": "D",
        "marker_style": "top",
        "label": "$2.0 \\leq y_{\\mathrm{b}} < 2.5$\n$0.0 \\leq y{*} < 0.5$",
        "zpt_binning": edge_binning,
    },
    "YB_00_05_YS_05_10": {
        "ys_range": (0.5, 1.0),
        "yb_range": (0.0, 0.5),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 7,
        "shift": 1.9,
        "color": "#2659a2",
        "marker": "s",
        "marker_style": "full",
        "label": "$0.0 \\leq y_{\\mathrm{b}} < 0.5$\n$0.5 \\leq y{*} < 1.0$",
        "zpt_binning": central_binning,
    },
    "YB_05_10_YS_05_10": {
        "ys_range": (0.5, 1.0),
        "yb_range": (0.5, 1.0),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 6,
        "shift": 1.7,
        "color": "#3e98b2",
        "marker": "s",
        "marker_style": "left",
        "label": "$0.5 \\leq y_{\\mathrm{b}} < 1.0$\n$0.5 \\leq y{*} < 1.0$",
        "zpt_binning": central_binning,
    },
    "YB_10_15_YS_05_10": {
        "ys_range": (0.5, 1.0),
        "yb_range": (1.0, 1.5),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 5,
        "shift": 1.5,
        "color": "#67b4cb",
        "marker": "s",
        "marker_style": "right",
        "label": "$1.0 \\leq y_{\\mathrm{b}} < 1.5$\n$0.5 \\leq y{*} < 1.0$",
        "zpt_binning": central_binning,
    },
    "YB_15_20_YS_05_10": {
        "ys_range": (0.5, 1.0),
        "yb_range": (1.5, 2.0),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 5,
        "shift": 1.3,
        "color": "#a0eaff",
        "marker": "s",
        "marker_style": "bottom",
        "label": "$1.5 \\leq y_{\\mathrm{b}} < 2.0$\n$0.5 \\leq y{*} < 1.0$",
        "zpt_binning": edge_binning,
    },
    "YB_00_05_YS_10_15": {
        "ys_range": (1.0, 1.5),
        "yb_range": (0.0, 0.5),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 3,
        "shift": 0.8,
        "color": "#139913",
        "marker": "v",
        "marker_style": "full",
        "label": "$0.0 \\leq y_{\\mathrm{b}} < 0.5$\n$1.0 \\leq y{*} < 1.5$",
        "zpt_binning": central_binning,
    },
    "YB_05_10_YS_10_15": {
        "ys_range": (1.0, 1.5),
        "yb_range": (0.5, 1.0),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 2,
        "shift": 0.6,
        "color": "#51c230",
        "marker": "v",
        "marker_style": "left",
        "label": "$0.5 \\leq y_{\\mathrm{b}} < 1.0$\n$1.0 \\leq y{*} < 1.5$",
        "zpt_binning": central_binning,
    },
    "YB_10_15_YS_10_15": {
        "ys_range": (1.0, 1.5),
        "yb_range": (1.0, 1.5),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 2,
        "shift": 0.4,
        "color": "#8def56",
        "marker": "v",
        "marker_style": "right",
        "label": "$1.0 \\leq y_{\\mathrm{b}} < 1.5$\n$1.0 \\leq y{*} < 1.5$",
        "zpt_binning": edge_binning,
    },
    "YB_00_05_YS_15_20": {
        "ys_range": (1.5, 2.0),
        "yb_range": (0.0, 0.5),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 1,
        "shift": -0.1,
        "color": "#c55200",
        "marker": "^",
        "marker_style": "full",
        "label": "$0.0 \\leq y_{\\mathrm{b}} < 0.5$\n$1.5 \\leq y{*} < 2.0$",
        "zpt_binning": edge_binning,
    },
    "YB_05_10_YS_15_20": {
        "ys_range": (1.5, 2.0),
        "yb_range": (0.5, 1.0),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 1,
        "shift": -0.3,
        "color": "#ffa500",
        "marker": "^",
        "marker_style": "left",
        "label": "$0.5 \\leq y_{\\mathrm{b}} < 1.0$\n$1.5 \\leq y{*} < 2.0$",
        "zpt_binning": edge_binning,
    },
    "YB_00_05_YS_20_25": {
        "ys_range": (2.0, 2.5),
        "yb_range": (0.0, 0.5),
        "pt_range": (25, 1000),
        "area": 0.25,
        "stagger": 0,
        "shift": -0.8,
        "color": "#a11313",
        "marker": "*",
        "marker_style": "full",
        "label": "$0.0 \\leq y_{\\mathrm{b}} < 0.5$\n$2.0 \\leq y{*} < 2.5$",
        "zpt_binning": extreme_binning,
    },
}

ybys_bins_inclusive = deepcopy(ybys_bins)
ybys_bins_inclusive.update(
    {
        "inclusive": {
            "ys_range": (0.0, 2.5),
            "yb_range": (0.0, 2.5),
            "pt_range": (25, 1000),
            "label": "$y_{\\mathrm{b}}$ inclusive\n$y{*}$ inclusive",
        }
    }
)
