import os
from enum import Enum, auto
from subprocess import PIPE

from law.logger import get_logger
from law.util import interruptable_popen

logger = get_logger(__name__)


source_env = dict()
for var in (
    "ZJET_REMOTE",
    "CMSSW_DIR",
    "X509_USER_PROXY",
    "HOME",
    "USER",
    "ANALYSIS_SOFTWARE",
):
    val = os.environ.get(var, None)
    if val:
        source_env[var] = val


def convert_env_to_dict(env):
    my_env = {}
    for line in env.splitlines():
        if line.find(" ") < 0:
            try:
                key, value = line.split("=", 1)
                my_env[key] = value
            except ValueError:
                pass
    return my_env


def set_environment_variables(ini_script_path):
    code, out, error = interruptable_popen(
        ". %s; env" % ini_script_path,
        shell=True,
        stdout=PIPE,
        stderr=PIPE,
        env=source_env,
    )
    if code != 0:
        raise RuntimeError(
            f"Sourcing {ini_script_path} failed with error code {code}!\n\n"
            f"Output:\n{out}\n"
            f"Error:\n{error}\n"
        )
    my_env = convert_env_to_dict(out)
    return my_env


kappa_env = set_environment_variables(
    os.path.expandvars("$ANALYSIS_BASE/scripts/ini_kappa.sh")
)

karma_env = set_environment_variables(
    os.path.expandvars("$ANALYSIS_BASE/scripts/ini_karma.sh")
)

excalibur_env = set_environment_variables(
    os.path.expandvars("$ANALYSIS_BASE/scripts/ini_excalibur.sh")
)


def run_command(command, *args, **kwargs):
    command_str = " ".join(command)
    logger.info(f'Running command: "{command_str}"')
    code, out, error = interruptable_popen(
        command, *args, stdout=PIPE, stderr=PIPE, **kwargs
    )
    if code != 0:
        raise RuntimeError(
            f"Command {command_str} failed with error code {code}!\n\n"
            f"Output:\n{out}\n"
            f"Error:\n{error}\n"
        )
    logger.debug(f"{command_str.split()[0]} ran successfully.\nOutput:\n{out}\n")
    return code, out, error


class DY_Sets(Enum):
    DYJetsToLL = auto()
    DYJetsToLL_0J = auto()
    DYJetsToLL_1J = auto()
    DYJetsToLL_2J = auto()
    DYJetsToLL_XJ = auto()
    DYJetsToLL_PtZ_0To50 = auto()
    DYJetsToLL_PtZ_50To100 = auto()
    DYJetsToLL_PtZ_100To250 = auto()
    DYJetsToLL_PtZ_250To400 = auto()
    DYJetsToLL_PtZ_400To650 = auto()
    DYJetsToLL_PtZ_650ToInf = auto()
    DYJetsToLL_PtZ = auto()
    DYJetsToLL_PtZ_combined = auto()
    DYJetsToLL_combined = auto()
    DYJetsToLL_MLM = auto()
    DYJetsToLL_MiNNLO = auto()
    DYJetsToLL_herwig = auto()


bg_names = ["WW", "ST", "WZ", "ZZ", "TTbar"]


data_periods_ul = ("2016ul", "2017ul", "2018ul")


jec_uncertainty_sources = [
    "Total",
    "AbsoluteMPFBias",
    "AbsoluteScale",
    "AbsoluteStat",
    "FlavorQCD",
    "Fragmentation",
    "PileUpDataMC",
    "PileUpPtBB",
    "PileUpPtEC1",
    "PileUpPtEC2",
    "PileUpPtHF",
    "PileUpPtRef",
    "RelativeFSR",
    "RelativeJEREC1",
    "RelativeJEREC2",
    "RelativeJERHF",
    "RelativePtBB",
    "RelativePtEC1",
    "RelativePtEC2",
    "RelativePtHF",
    "RelativeBal",
    "RelativeSample",
    "RelativeStatEC",
    "RelativeStatFSR",
    "RelativeStatHF",
    "SinglePionECAL",
    "SinglePionHCAL",
    "TimePtEta",
    "SubTotalPileUp",
    "SubTotalRelative",
    "SubTotalPt",
    "SubTotalScale",
]


class Energy_Shifts(Enum):
    NONE = auto()
    JEC_UP = auto()
    JEC_DOWN = auto()
    JER_UP = auto()
    JER_DOWN = auto()
    # ROCCOR_UP = auto()
    # ROCCOR_DOWN = auto()


def get_energy_shift_ext(shift, *args):
    args = list(args)
    if shift is Energy_Shifts.NONE:
        return ""
    if "Total" in args:
        args.remove("Total")
    ret = "_".join([shift.name] + args)
    return f"_{ret}"
