import os

import law
import luigi
from analysis.framework.dataset_map import cross_sections, dataset_map
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig, ExternalCMSSWTask
from analysis.tasks.Excalibur.CreatePileupWeights import (
    CreatePileupWeights,
    PileupShift,
)
from analysis.tasks.Excalibur.GetGeneratorWeights import GetGeneratorWeights
from analysis.tasks.Skimming import SkimDataset
from luigi.util import inherits

DY_SETS = ("DYJetsToLL", "DYJetsToLL_0J", "DYJetsToLL_1J", "DYJetsToLL_2J")


class ExcaliburMCConfigTemplate(ExternalCMSSWTask):
    jet_type = luigi.Parameter(default="AK4")

    def output(self):
        jet_ext = "_Puppi" if "Puppi" in self.jet_type else ""
        return law.LocalFileTarget(
            f"{self.cmssw_path}/src/Excalibur/cfg/excalibur/zjet{self.data_period}/"
            f"mc_{self.data_period}_singlemuon{jet_ext}.py.jinja2"
        )


@inherits(ExcaliburConfig)
class CreateExcaliburMCConfig(AnalysisTask):
    jet_type = luigi.Parameter(default="AK4")

    def requires(self):
        _ret = {
            "Template": ExcaliburMCConfigTemplate.req(self),
            "PileUpWeights": CreatePileupWeights.req(
                self, pileup_shift=PileupShift.nominal
            ),
            "PileUpWeights_up": CreatePileupWeights.req(
                self, pileup_shift=PileupShift.up
            ),
            "PileUpWeights_down": CreatePileupWeights.req(
                self, pileup_shift=PileupShift.down
            ),
            "GetGeneratorWeights": GetGeneratorWeights.req(self),
            "Skim": SkimDataset.req(self),
        }
        return _ret

    def output(self):
        jet_ext = "_Puppi" if "Puppi" in self.jet_type else ""
        return law.LocalFileTarget(
            os.path.join(
                f"{self.cmssw_path}/src/Excalibur/cfg/excalibur",
                f"zjet{self.data_period}",
                self.excalibur_config_version,
                f"{self.data_period}_{self.dataset}{jet_ext}.py",
            )
        )

    def run(self):
        from jinja2 import Template

        dataset_vars = dataset_map[self.data_period]["common_mc"]
        dataset_vars.update(dataset_map[self.data_period][self.dataset])

        dataset_wildcards = [
            col.dir.uri() + "/*"
            for col in self.input()["Skim"]["collection"].collections
        ]
        variables = dict(
            year=str(self.data_period).strip("ulUL_APVprepostVFP"),
            dataset_files=dataset_wildcards,
            jet_radius=self.jet_radius,
            jec=dataset_vars["jec_version"],
            jer=dataset_vars["jer_version"],
            roccor=dataset_vars["roccor"],
            pileup_weight_file=os.path.abspath(self.input()["PileUpWeights"].abspath),
            pileup_weight_file_up=os.path.abspath(
                self.input()["PileUpWeights_up"].abspath
            ),
            pileup_weight_file_down=os.path.abspath(
                self.input()["PileUpWeights_down"].abspath
            ),
            scale_weights=cross_sections[self.dataset]["scale_weights"],
        )
        variables.update(cross_sections[self.dataset])
        variables["gen_weights_scale"] = [
            f"genWeight_{_id}" for _id in cross_sections[self.dataset]["scale_weights"]
        ]
        variables["gen_weights_pdf"] = [
            f"genWeight_{_id}" for _id in cross_sections[self.dataset]["pdf_weights"]
        ]
        variables["gen_weights_ps"] = [
            f"genWeight_{_id}" for _id in cross_sections[self.dataset]["ps_weights"]
        ]

        gen_weights = self.input()["GetGeneratorWeights"].load()
        variables.update(gen_weights)

        # create file from jinja template
        with open(self.input()["Template"].abspath) as file_:
            template = Template(file_.read())
        config = template.render(variables)
        with self.output().localize("w") as tmp_file:
            with tmp_file.open("w") as g:
                g.write(config)
        return 0
