from enum import Enum

import law
import luigi
from analysis.framework.dataset_map import dataset_map
from analysis.framework.tasks import AnalysisTask
from analysis.framework.utils import excalibur_env, run_command
from analysis.tasks.Skimming import SkimDataset
from law.decorator import safe_output


class PileupShift(Enum):
    # 4.6% shifts for pileup estimation
    up = 1.0 + 0.046
    down = 1.0 - 0.046
    nominal = 1.0


class CreatePileupWeights(AnalysisTask):
    pileup_shift = luigi.EnumParameter(
        default=PileupShift.nominal,
        enum=PileupShift,
        description="Pileup uncertainty for MC weights. Choose from nominal, up or down.",
    )

    def requires(self):
        return SkimDataset.req(self)

    def output(self):
        path = f"{self.cmssw_path}/src/Excalibur/data/pileup/mc_weights/{self.data_period}/{self.dataset}_{self.pileup_shift.name}.root"
        return law.LocalFileTarget(path)

    @safe_output
    def run(self):
        # Good run jsons from: https://twiki.cern.ch/twiki/bin/viewauth/CMS/TWikiLUM
        good_run_json = dataset_map[self.data_period]["common_mc"]["good_run_json"]
        # Pileup distribution from: lxplus at CERN:
        # /afs/cern.ch/cms/CAF/CMSCOMM/COMM_DQM/certification/Collisions17/13TeV/PileUp/
        data_pu_dist = dataset_map[self.data_period]["common_mc"][
            "data_pu_distribution"
        ]
        input_files = self.input()["collection"].uri()
        tmp_workdir = law.LocalDirectoryTarget(is_tmp=True)
        tmp_workdir.touch()

        # run tool to get generator weight
        pu_value = 69.2 * self.pileup_shift.value
        command = (
            [
                "puWeightCalc.py",
                f"{self.cmssw_path}/src/Excalibur/{good_run_json}",
            ]
            + input_files
            + [
                "-x",
                str(pu_value),
                "-i",
                f"{self.cmssw_path}/src/Excalibur/{data_pu_dist}",
                "-o",
                f"{self.output().abspath}",
            ]
        )
        run_command(
            command,
            env=excalibur_env,
            cwd=tmp_workdir.abspath,
        )


class AllPUWeights(law.WrapperTask):
    data_period = luigi.Parameter()

    def requires(self):
        ret = []
        for dataset in dataset_map[self.data_period]:
            if (
                dataset
                in (
                    "common_mc",
                    "common_data",
                    "DYJetsToMuMu",
                    "DYJetsToTauTau",
                )
                or "data" in dataset
            ):
                continue
            for shift in PileupShift:
                ret.append(
                    CreatePileupWeights.req(self, dataset=dataset, pileup_shift=shift)
                )
        return ret
