import concurrent.futures
import json
from copy import deepcopy

import law
import luigi
from analysis.framework.dataset_map import dataset_map
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig
from analysis.framework.utils import (
    Energy_Shifts,
    excalibur_env,
    get_energy_shift_ext,
    jec_uncertainty_sources,
    run_command,
)
from analysis.tasks.Excalibur import CreateExcaliburDataConfig, CreateExcaliburMCConfig
from analysis.tasks.Remote import HTCondorWorkflow
from analysis.tasks.Skimming import SkimDataset
from law.decorator import localize
from law.logger import get_logger
from luigi.util import inherits

logger = get_logger(__name__)


@inherits(ExcaliburConfig)
class ExcaliburJsonConfig(AnalysisTask):
    jet_type = luigi.Parameter(default="AK4")

    def requires(self):
        if "data" in self.dataset:
            return CreateExcaliburDataConfig.req(self)
        else:
            return CreateExcaliburMCConfig.req(self)

    def output(self):
        jet_ext = "_Puppi" if "Puppi" in self.jet_type else ""
        return self.remote_target(
            self.excalibur_config_version,
            f"{self.dataset}{jet_ext}.json",
        )

    def run(self):
        command = ["excalibur.py", "-c", self.input().abspath]
        run_command(command, env=excalibur_env, cwd=f"{self.cmssw_path}/src/Excalibur")
        out_json = law.LocalFileTarget(is_tmp="json")
        with open(f"{self.input().abspath}.json", "r") as json_cfg:
            tmp = json_cfg.read()
            tmp = tmp.replace(self.cmssw_path, "$CMSSW_DIR")
            out_json.open("w").write(tmp)
        self.output().move_from_local(out_json)


@inherits(ExcaliburConfig)
class ExcaliburWorkflow(AnalysisTask, HTCondorWorkflow, law.LocalWorkflow):
    jet_type = luigi.Parameter(default="AK4")
    files_per_job = luigi.IntParameter(
        default=2,
        description="Number of files analyzed per job.",
    )
    exclude_params_req = {"files_per_job"}

    def workflow_requires(self):
        return {"json_cfg": ExcaliburJsonConfig.req(self)}

    def output(self):
        jet_ext = "_Puppi" if "Puppi" in self.jet_type else ""
        return self.remote_target(
            self.excalibur_config_version,
            f"{self.dataset}{jet_ext}",
            f"job_{self.branch}.root",
        )

    def create_branch_map(self):
        return SkimDataset.req(self).get_all_branch_chunks(self.files_per_job)

    def requires(self):
        return {"skim": SkimDataset.req(self, branch=-1, branches=self.branch_data)}

    @localize(input=True, output=True)
    def run(self):
        cwd = law.LocalDirectoryTarget(is_tmp="excalibur")
        cwd.touch()
        with self.workflow_input()["json_cfg"].localize() as json_cfg:
            cfg = json_cfg.open(mode="r").read()
        cfg = cfg.replace("$CMSSW_DIR", self.cmssw_path)
        cfg_dict = json.loads(cfg)
        cfg_dict["InputFiles"] = self.input()["skim"]["collection"].uri()

        if "data" in self.dataset:
            raise NotImplementedError(
                "Implementing data should be easy, just skip all shifts"
            )

        # build list of all shifts to run
        shift_list = []
        for energy_shift in Energy_Shifts:
            if energy_shift in (Energy_Shifts.JEC_UP, Energy_Shifts.JEC_DOWN):
                for jec in jec_uncertainty_sources:
                    shift_list.append((energy_shift, jec))
            else:
                shift_list.append((energy_shift, "Total"))

        # execute excalibur for each energy shift in parallel up to the requested number of CPUs
        futures = []
        output_files = []
        # Use ThreadPoolExecutor to run the commands in parallel
        # Overbook with one more thread, as Excalibur only has 80% CPU usage
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.htcondor_cpus + 1
        ) as executor:
            for energy_shift, jec in shift_list:
                shift_json = law.LocalFileTarget(
                    f"{cwd.abspath}/{energy_shift.name}_{jec}.json",
                )
                shift_root = law.LocalFileTarget(
                    f"{cwd.abspath}/{energy_shift.name}_{jec}.root"
                )
                output_files.append(shift_root)
                cfg_shift = deepcopy(cfg_dict)
                cfg_shift["OutputPath"] = shift_root.abspath
                cfg_shift["JetEnergyCorrectionUncertaintySource"] = jec
                if energy_shift == Energy_Shifts.NONE:
                    pass
                elif energy_shift == Energy_Shifts.JER_UP:
                    cfg_shift["JERShift"] = int(1)
                    cfg_shift["Pipelines"].pop("genzjetcuts_L1L2L3")
                    cfg_shift["Pipelines"]["zjetcuts_L1L2L3_JER_UP"] = cfg_shift[
                        "Pipelines"
                    ].pop("zjetcuts_L1L2L3")
                elif energy_shift == Energy_Shifts.JER_DOWN:
                    cfg_shift["JERShift"] = int(-1)
                    cfg_shift["Pipelines"].pop("genzjetcuts_L1L2L3")
                    cfg_shift["Pipelines"]["zjetcuts_L1L2L3_JER_DOWN"] = cfg_shift[
                        "Pipelines"
                    ].pop("zjetcuts_L1L2L3")
                elif energy_shift == Energy_Shifts.JEC_UP:
                    cfg_shift["JetEnergyCorrectionUncertaintyShift"] = 1.0
                    cfg_shift["Pipelines"].pop("genzjetcuts_L1L2L3")
                    cfg_shift["Pipelines"][f"zjetcuts_L1L2L3_JEC_{jec}_UP"] = cfg_shift[
                        "Pipelines"
                    ].pop("zjetcuts_L1L2L3")
                elif energy_shift == Energy_Shifts.JEC_DOWN:
                    cfg_shift["JetEnergyCorrectionUncertaintyShift"] = -1.0
                    cfg_shift["Pipelines"].pop("genzjetcuts_L1L2L3")
                    cfg_shift["Pipelines"][f"zjetcuts_L1L2L3_JEC_{jec}_DOWN"] = (
                        cfg_shift["Pipelines"].pop("zjetcuts_L1L2L3")
                    )
                else:
                    raise ValueError(f"Unknown energy shift: {energy_shift}")
                shift_json.dump(cfg_shift, indent=4)
                logger.info(f"Adding json_cfg: {shift_json.abspath}")
                future = executor.submit(
                    run_command,
                    ["excalibur", shift_json.abspath],
                    env=excalibur_env,
                    cwd=cwd.abspath,
                )
                futures.append(future)

            # Wait for all futures to complete and check for errors
            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()
                except RuntimeError as e:
                    raise e

        # Hadd all shifts and move to remote storage
        filelist = [x.abspath for x in output_files]
        command = ["hadd", "-f", self.output().abspath] + filelist
        self.output().touch()
        run_command(command)


@inherits(ExcaliburConfig)
class ExcaliburMerge(AnalysisTask):
    def requires(self):
        return ExcaliburWorkflow.req(self)

    def output(self):
        return self.remote_target(
            self.excalibur_config_version,
            self.dataset,
            f"{self.dataset}{get_energy_shift_ext(self.energy_shift, self.jec_uncertainty_source)}.root",
        )

    @localize(input=False, output=True)
    def run(self):
        command = ["pseudo_hadd.py", self.output().abspath] + self.input()[
            "collection"
        ].uri()
        run_command(command, env=excalibur_env)


class ExcaliburAll(law.WrapperTask):
    data_period = luigi.Parameter()
    excalibur_config_version = luigi.Parameter()
    jet_radius = luigi.Parameter(default="4")

    def requires(self):
        reqs = {}
        for dataset in dataset_map[self.data_period].keys():
            if "common" in dataset:
                continue
            reqs[dataset] = ExcaliburMerge.req(
                self,
                dataset=dataset,
            )
            if "DYJetsToLL" in dataset and dataset != "DYJetsToLL_M-10to50":
                for jec_uncertainty_source in jec_uncertainty_sources:
                    if jec_uncertainty_source == "Total":
                        for energy_shift in (
                            Energy_Shifts.JER_UP,
                            Energy_Shifts.JER_DOWN,
                        ):
                            reqs[
                                f"{dataset}{get_energy_shift_ext(energy_shift, jec_uncertainty_source)}"
                            ] = ExcaliburMerge.req(
                                self,
                                dataset=dataset,
                                energy_shift=energy_shift,
                            )
                    for energy_shift in (
                        Energy_Shifts.JEC_UP,
                        Energy_Shifts.JEC_DOWN,
                    ):
                        reqs[
                            f"{dataset}{get_energy_shift_ext(energy_shift, jec_uncertainty_source)}"
                        ] = ExcaliburMerge.req(
                            self,
                            dataset=dataset,
                            energy_shift=energy_shift,
                            jec_uncertainty_source=jec_uncertainty_source,
                        )
        return reqs
