import luigi

from analysis.framework.tasks import AnalysisTask
from analysis.framework.utils import run_command, karma_env, DY_Sets
from analysis.tasks.PostProcessing import Histograms, MergeMCShapeUncertainties


class PlotDoubleRatio(AnalysisTask):
    excalibur_config_version = luigi.Parameter()
    karma_config = luigi.Parameter(default="zjet_excalibur")

    selections = luigi.Parameter()
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)

    data_period_num = luigi.Parameter()
    data_period_den = luigi.Parameter()
    weight_mc = luigi.Parameter(default="totalWeight")
    weight_data = luigi.Parameter(default="weight")
    quantities = luigi.ListParameter()

    def requires(self):
        _req = dict()
        for abbr, dataset in zip(
            ("Data", "DY", "TT", "WZ", "ZZ", "ST", "WW"),
            ("data", self.mc_dataset.name, "TTbar", "WZ", "ZZ", "ST", "WW"),
        ):
            _req["Num" + abbr] = Histograms(
                splitting="ybys",
                data_period=self.data_period_num,
                dataset=dataset,
                task="DATAShape",
                excalibur_config_version=self.excalibur_config_version,
                karma_config=self.karma_config,
                selections=self.selections,
            )
            _req["Den" + abbr] = Histograms(
                splitting="ybys",
                data_period=self.data_period_den,
                dataset=dataset,
                task="DATAShape",
                excalibur_config_version=self.excalibur_config_version,
                karma_config=self.karma_config,
                selections=self.selections,
            )
        _req["NumUncertainties"] = MergeMCShapeUncertainties.req(
            self,
            data_period=self.data_period_num,
            dataset=self.mc_dataset.name,
            task="DATAShape",
            splitting="ybys",
        )
        _req["DenUncertainties"] = MergeMCShapeUncertainties.req(
            self,
            data_period=self.data_period_den,
            dataset=self.mc_dataset.name,
            task="DATAShape",
            splitting="ybys",
        )
        return _req

    def output(self):
        return self.local_plot_target(
            f"{self.excalibur_config_version}_{self.selections}_cuts",
            f"{self.data_period_num}-{self.data_period_den}/",
        )

    def run(self):
        command = (
            [
                "palisade.py",
                "task",
                self.karma_config,
                "plot_doubleratio_quantity",
                "--quantities",
            ]
            + list(self.quantities)
            + [
                "--data-num",
                "{}".format(self.input()["NumData"].abspath),
                "--data-den",
                "{}".format(self.input()["DenData"].abspath),
                "--mc-num",
                self.input()["NumDY"].abspath,
                self.input()["NumTT"].abspath,
                self.input()["NumWZ"].abspath,
                self.input()["NumZZ"].abspath,
                self.input()["NumST"].abspath,
                self.input()["NumWW"].abspath,
                "--mc-den",
                self.input()["DenDY"].abspath,
                self.input()["DenTT"].abspath,
                self.input()["DenWZ"].abspath,
                self.input()["DenZZ"].abspath,
                self.input()["DenST"].abspath,
                self.input()["DenWW"].abspath,
                "--unc-num",
                self.input()["NumUncertainties"].abspath,
                "--unc-de",
                self.input()["DenUncertainties"].abspath,
                "--weight-data",
                self.weight_data,
                "--weight-mc",
                self.weight_mc,
                "--data-period-num",
                self.data_period_num.replace("ul", ""),
                "--data-period-den",
                self.data_period_den.replace("ul", ""),
                "--output-dir",
                self.output().abspath,
                "--output-format",
                "{quantity}.pdf",
                "--inclusive-only",
            ]
        )
        print(f"command: {command}")
        try:
            run_command(command, env=karma_env, cwd=self.cmssw_path)
        except RuntimeError as e:
            self.output().remove()
            raise e
