import matplotlib.pyplot as plt
import numpy as np
import uproot

import law
import luigi
import mplhep as hep
from analysis.framework.quantities import QUANTITIES
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig
from analysis.tasks.PostProcessing import Histograms
from law.contrib.matplotlib import MatplotlibFormatter
from luigi.util import inherits


@inherits(ExcaliburConfig)
class PlotMigrationMatrix2(AnalysisTask):
    quantities = luigi.ListParameter(default=["binIndexzpt"])
    weight_name = luigi.Parameter()
    forward_smearing = luigi.BoolParameter()
    selection_suffix = luigi.Parameter()

    def requires(self):
        if self.forward_smearing:
            raise NotImplementedError()
        else:
            return {
                "migration": Histograms.req(
                    self,
                    splitting="ybys",
                    task="CreateResponseMatrix",
                    selections="response{}".format(self.selection_suffix),
                ),
            }

    def output(self):
        name = self.__class__.__name__.replace("Plot", "")
        filelist = {}
        for quantity in self.quantities:
            filelist[quantity] = law.LocalFileTarget(
                self.local_path(
                    f"{self.excalibur_config_version}_{self.selection_suffix}_cuts",
                    f"{self.dataset}_{self.weight_name}",
                    f"migration_{quantity}.pdf",
                    name=name,
                ),
                fs="local_plots",
            )
        return law.SiblingFileCollection(filelist)

    def run(self):
        self.output().dir.touch()
        for quantity in self.quantities:
            hep.style.use("CMS")
            fig, ax = plt.subplots()
            hep.cms.label(label="Preliminary", loc=0, ax=ax)

            matrix, xbins, ybins = uproot.open(self.input()["migration"].abspath)[
                f"inclusive/{quantity}/h2d_{QUANTITIES[quantity]['gen_name']}_{self.weight_name}"
            ].to_numpy()
            # normalize to gen (x-Axis)
            matrix /= matrix.sum(axis=1, keepdims=True)
            matrix[matrix == 0] = np.nan  # drop empty bins

            hep.hist2dplot(
                matrix,
                xbins=xbins,
                ybins=ybins,
                # cmin=1e-4,
                cmax=1.0,
                cbarextend=True,
                flow=None,
                # norm="log",
                ax=ax,
            )
            ax.set_xlabel(QUANTITIES[quantity]["gen_label"])
            ax.set_ylabel(QUANTITIES[quantity]["reco_label"])
            plt.tight_layout()
            MatplotlibFormatter.dump(self.output().targets[quantity], fig)
