import matplotlib.pyplot as plt
import numpy as np
import uproot

import law
import luigi
import mplhep as hep
from analysis.framework.quantities import QUANTITIES, lumis, ybys_bins_inclusive
from analysis.framework.tasks import AnalysisTask, LumberjackConfig
from analysis.framework.utils import DY_Sets, bg_names, karma_env, run_command
from analysis.tasks.PostProcessing import Histograms, MergeMCShapeUncertainties
from analysis.tools.plotting.axes_ticker import set_ticklabels
from analysis.tools.plotting.legend_helpers import inline_label
from law.contrib.matplotlib import MatplotlibFormatter
from law.workflow.local import LocalWorkflow
from luigi.util import inherits


@inherits(LumberjackConfig)
class PlotSignalBackground(AnalysisTask, LocalWorkflow):
    # allow outputs in nested directory structure
    output_collection_cls = law.NestedSiblingFileCollection

    dataset = luigi.Parameter(default="data")
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)

    weight_data = luigi.Parameter(default="weight")
    weight_mc = luigi.Parameter(default="totalWeight")

    ratio_range = luigi.TupleParameter(default=(0.75, 1.25))
    scale_to_data = luigi.BoolParameter(default="False")
    quantities = luigi.ListParameter(default=QUANTITIES.keys())

    def create_branch_map(self):
        branches = {i: q for i, q in enumerate(self.quantities)}
        return branches

    def workflow_requires(self):
        req = {
            "uncs": MergeMCShapeUncertainties.req(
                self, dataset=self.mc_dataset.name, task="DATAShape", splitting="ybys"
            ),
            "data": Histograms.req(
                self,
                splitting="ybys",
                data_period=self.data_period,
                dataset=self.dataset,
                task="DATAShape",
            ),
            "DY": Histograms.req(
                self,
                splitting="ybys",
                dataset=self.mc_dataset.name,
                task="DATAShape",
            ),
        }
        for bg in bg_names:
            req[bg] = Histograms.req(
                self,
                dataset=bg,
                splitting="ybys",
                task="DATAShape",
            )
        return req

    def output(self):
        plots = {}
        for ybys in ybys_bins_inclusive:
            if self.branch_data not in QUANTITIES:
                raise ValueError(f"Unknwon quantity {self.branch_data}!")
            plots[f"{ybys}/{self.branch_data}"] = self.local_plot_target(
                f"{self.excalibur_config_version}_{self.selections}",
                f"{self.dataset}-{self.weight_data}_{self.mc_dataset.name}-{self.weight_mc}",
                f"{ybys}",
                f"{self.branch_data}.pdf",
            )
        return law.NestedSiblingFileCollection(plots)

    def plot_quantity(self, ybys, quantity, q_dict):
        hep.style.use("CMS")
        fig, (axmain, axratio) = plt.subplots(
            2, 1, sharex=True, gridspec_kw={"height_ratios": [3, 1]}
        )
        lumi = lumis[str(self.data_period).replace("ul", "")]
        hep.cms.label(
            label="Preliminary", data=True, lumi=f"{lumi:.3g}", loc=0, ax=axmain
        )
        inline_label(axmain, ybys_bins_inclusive[ybys]["label"])
        data, bins = self.data_file[
            f"{ybys}/h_{quantity}_{self.weight_data}"
        ].to_numpy()
        data_errors = self.data_file[f"{ybys}/h_{quantity}_{self.weight_data}"].errors()
        hep.histplot(
            data,
            bins,
            yerr=data_errors,
            label="Data",
            color="black",
            histtype="errorbar",
            binwnorm=1,
            ax=axmain,
        )
        mc_labels = bg_names + ["DY"]
        mc_hists = np.array(
            [
                self.mc_files[mc][f"{ybys}/h_{quantity}_{self.weight_mc}"].to_numpy()[0]
                for mc in mc_labels
            ]
        )
        mc_stack = np.sum(mc_hists, axis=0)
        if np.any(mc_stack < 0):
            print(f"MC stack containes negative events for {ybys}/{quantity}!")
            print(mc_stack)

        # ignore division by 0 errors when plotting ratios
        with np.errstate(divide="ignore", invalid="ignore"):
            hep.histplot(
                mc_hists,
                bins,
                label=mc_labels,
                histtype="fill",
                stack=True,
                binwnorm=1,
                ax=axmain,
            )
            hep.histplot(
                data / mc_stack,
                bins,
                yerr=data_errors / np.abs(mc_stack),
                histtype="errorbar",
                color="black",
                ax=axratio,
            )
            mc_unc_total = np.abs(
                [
                    self.unc_file[f"{ybys}/h_{quantity}_total_down"].to_numpy()[0],
                    self.unc_file[f"{ybys}/h_{quantity}_total_up"].to_numpy()[0],
                ]
            )
            hep.histplot(
                mc_stack,
                bins,
                yerr=mc_unc_total / mc_stack,
                color="gray",
                histtype="band",
                alpha=0.5,
                ax=axratio,
            )
            hep.histplot(
                np.ones_like(mc_stack),
                bins,
                yerr=mc_unc_total / mc_stack,
                label="MC unc.",
                color="gray",
                histtype="band",
                alpha=0.5,
                ax=axratio,
            )
        axratio.hlines(1, *q_dict["range"], color="black", linestyle="--")

        axmain.set_xscale(q_dict.get("scale", "linear"))
        axmain.set_yscale("log")
        # axmain.set_ylim(1e-2, 1e6)
        axmain.set_xlim(q_dict["range"])
        axmain.set_ylabel("Events / GeV")
        axratio.set_xlabel(q_dict["label"])
        axratio.set_ylabel("Data / MC")
        axratio.set_ylim(self.ratio_range)

        set_ticklabels(axmain, suppress_x_ticklabels=True)
        set_ticklabels(axratio, suppress_x_ticklabels=False)

        mc_labels.reverse()
        handles, labels = hep.sort_legend(axmain, ["Data"] + mc_labels)
        axmain.legend(handles, labels, ncol=2, loc="upper right")
        axratio.legend(ncol=1, loc="upper left")
        axmain = hep.yscale_legend(axmain)
        fig.tight_layout()
        MatplotlibFormatter.dump(self.output().targets[f"{ybys}/{quantity}"], fig)
        plt.close(fig)

    def run(self):
        for collection in self.output().collections:
            collection.dir.touch()
        self.data_file = uproot.open(self.workflow_input()["data"].abspath)
        self.unc_file = uproot.open(self.workflow_input()["uncs"].abspath)
        self.mc_files = {
            bg: uproot.open(self.workflow_input()[bg].abspath) for bg in bg_names
        }
        self.mc_files["DY"] = uproot.open(self.workflow_input()["DY"].abspath)
        quantity = self.branch_data
        for ybys in ybys_bins_inclusive:
            if self.output().targets[f"{ybys}/{quantity}"].exists():
                continue
            self.plot_quantity(ybys, quantity, q_dict=QUANTITIES[quantity])


@inherits(LumberjackConfig)
class PlotMCShapeUncertainties(AnalysisTask):
    dataset = luigi.Parameter(default=DY_Sets.DYJetsToLL_combined.name)
    input_weight = luigi.Parameter(default="weight")

    plot_range = luigi.Parameter(default="m0.25,0.35")
    quantities = luigi.ListParameter()

    def requires(self):
        unc = MergeMCShapeUncertainties.req(self)
        req = {
            "DY": unc.requires()["DY"],
            "unc": unc,
        }
        return req

    def output(self):
        return self.local_plot_target(
            f"{self.excalibur_config_version}_{self.selections}_cuts",
            f"{self.dataset}-{self.input_weight}/",
        )

    def run(self):
        self.output().touch()
        command = (
            [
                "palisade.py",
                "task",
                "zjet_excalibur",
                "plot_uncertainties_mc_shapes",
                "--data-period",
                self.data_period.replace("ul", ""),
                "--quantities",
            ]
            + list(self.quantities)
            + [
                "--filename-mc",
                self.input()["DY"].abspath,
                "--filename-unc",
                self.input()["unc"].abspath,
                "--weight-mc",
                self.input_weight,
                "--plot-range",
                self.plot_range,
                "--output-format",
                "unc_{ybys[name]}_{quantity[name]}.pdf",
                "--output-dir",
                self.output().abspath,
            ]
        )
        try:
            run_command(command, env=karma_env)
        except RuntimeError as e:
            self.output().remove()
            raise e
