import matplotlib.pyplot as plt
import numpy as np
import uproot

import law
import luigi
import mplhep as hep
from analysis.framework.quantities import QUANTITIES, ybys_bins
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig
from analysis.framework.utils import DY_Sets
from analysis.tasks.PostProcessing import MergeUncertainties
from analysis.tools.plotting.axes_ticker import set_ticklabels
from analysis.tools.plotting.legend_helpers import inline_label
from law.contrib.matplotlib import MatplotlibFormatter
from luigi.util import inherits


@inherits(ExcaliburConfig)
class PlotUncertainties(AnalysisTask):
    dataset = luigi.Parameter(
        default="data"
        # default=DY_Sets.DYJetsToLL_combined.name
    )  # overwrite default from ExcaliburConfig
    input_is_mc = luigi.BoolParameter(default=False)
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    unfolding_weight = luigi.Parameter(default="totalWeight")
    gen_weight = luigi.Parameter("weight")

    bkg_variation_sfs = luigi.ListParameter(default=[1.5, 0.5])

    selection_suffix = luigi.Parameter()
    plot_range = luigi.Parameter(default="m0.25,0.35")

    def requires(self):
        unc = MergeUncertainties.req(self)
        req = {
            "data": unc.requires()["unf"],
            "unc": unc,
            # 'gen': Histograms(
            #     splitting="genybys", data_period=self.data_period,
            #     dataset=self.mc_dataset,
            #     task="MCShape",
            #     selections="gen{}".format(self.selection_suffix),
            #     excalibur_config_version=self.excalibur_config_version
            #     ),
        }
        return req

    def output(self):
        filelist = {}
        for ybys_bin in ybys_bins:
            splits = {}
            for split in ("total", "jet", "lum", "sf", "stat"):
                splits[split] = law.LocalFileTarget(
                    self.local_path(
                        f"{self.excalibur_config_version}_{self.selection_suffix}_cuts",
                        f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset}-{self.unfolding_weight}",
                        f"{ybys_bin}_{split}.pdf",
                    ),
                    fs="local_plots",
                )
            filelist[ybys_bin] = splits
        return law.SiblingFileCollection(filelist)

    def run(self):
        data_file = uproot.open(self.input()["data"].abspath)
        unc_file = uproot.open(self.input()["unc"].abspath)
        uncertainties = {
            "total": dict(label="Total", color="black", linewidth=3),
            "stat": dict(color="#999999", label="Stat. unc."),
            "unf": dict(label="MC stat.", color="#377eb8", linestyle="dashdot"),
            "smooth_jec_Total_calc": dict(color="#e41a1c", label="JES"),
            "jer": dict(
                color="#a65628", linestyle=(0, (3, 1, 1, 1, 1, 1)), label="JER"
            ),
            "puJetID": dict(color="#ff7f00", linestyle="dotted", label="PUJetID"),
            "muon": dict(color="#ffff33", linestyle="dashed", label="Muon SFs"),
            "prefiring": dict(color="#4daf4a", label="L1 Prefiring"),
            "bkgsf": dict(color="#f781bf", label="Background"),
            "lumi": dict(color="#984ea3", label="Lumi"),
            "PileUp": dict(color="#47429d", linestyle="dotted", label="PileUp"),
        }
        self.output().dir.touch()
        hep.style.use("CMS")
        for ybys, _d in ybys_bins.items():
            for split in ("total", "jet", "lum", "sf", "stat"):
                fig, ax = plt.subplots()
                hep.cms.label(
                    label="Preliminary",
                    data=not self.input_is_mc,
                    lumi=138,
                    loc=0,
                    ax=ax,
                )
                inline_label(ax, _d["label"])
                ax.set_xscale("log")
                ax.set_xlim(25, 1000)
                ax.set_ylim(-0.2, 0.4)
                ax.hlines(0, 25, 1000, color="black", linestyle="dashed")
                data, bins = data_file[
                    f"{ybys}/h_zpt_{self.unfolding_weight}"
                ].to_numpy()
                for unc, unc_dict in uncertainties.items():
                    if split == "jet" and unc not in ("smooth_jec_Total_calc", "jer"):
                        continue
                    elif split == "lum" and unc not in ("lumi", "bkgsf", "PileUp"):
                        continue
                    elif split == "sf" and unc not in ("muon", "prefiring"):
                        continue
                    elif split == "stat" and unc not in ("stat", "unf"):
                        continue
                    up, _ = unc_file[f"{ybys}/h_zpt_{unc}_up"].to_numpy()
                    down, _ = unc_file[f"{ybys}/h_zpt_{unc}_down"].to_numpy()
                    ax.stairs(
                        up / data,
                        edges=bins,
                        label=unc_dict["label"],
                        color=unc_dict["color"],
                        linestyle=unc_dict.get("linestyle", "-"),
                        linewidth=unc_dict.get("linewidth", 2),
                        baseline=None,
                    )
                    ax.stairs(
                        down / data,
                        edges=bins,
                        color=unc_dict["color"],
                        linestyle=unc_dict.get("linestyle", "-"),
                        linewidth=unc_dict.get("linewidth", 2),
                        baseline=None,
                    )
                ax.set_ylabel("Relative uncertainty")
                ax.set_xlabel(QUANTITIES["zpt"]["label"])
                set_ticklabels(ax)
                ax.legend(ncol=2)
                fig.tight_layout()
                MatplotlibFormatter.dump(self.output().targets[ybys][split], fig)
                plt.close(fig)


@inherits(ExcaliburConfig)
class PlotCorrelationMatrix(AnalysisTask):
    dataset = luigi.Parameter(
        default="data"
        # default=DY_Sets.DYJetsToLL_combined.name
    )  # overwrite default from ExcaliburConfig
    input_is_mc = luigi.BoolParameter(default=False)
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    unfolding_weight = luigi.Parameter(default="totalWeight")
    gen_weight = luigi.Parameter("weight")

    bkg_variation_sfs = luigi.ListParameter(default=[1.5, 0.5])

    selection_suffix = luigi.Parameter()
    plot_range = luigi.Parameter(default="m0.25,0.35")

    def requires(self):
        unc = MergeUncertainties.req(self)
        reravel = unc.requires()["unf"]
        unfolded = reravel.requires()["Unfold"]
        return unfolded

    def output(self):
        return law.LocalFileTarget(
            self.local_path(
                f"{self.excalibur_config_version}_{self.selection_suffix}_cuts",
                f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset}-{self.unfolding_weight}",
                "correlation_matrix.pdf",
                name="CorrelationMatrix",
            ),
            fs="local_plots",
        )

    def run(self):
        quantity = "binIndexzpt"
        self.output().parent.touch()
        hep.style.use("CMS")
        fig, ax = plt.subplots()
        hep.cms.label(
            label="Preliminary", data=not self.input_is_mc, lumi=138, loc=0, ax=ax
        )
        covariance, xbins, ybins = uproot.open(self.input().abspath)[
            f"inclusive/h2d_{quantity}_{self.unfolding_weight}_stat_unc"
        ].to_numpy()

        v = np.sqrt(np.diag(covariance))
        outer_v = np.outer(v, v)
        correlation = covariance / outer_v
        correlation[covariance == 0] = 0

        hist2d = hep.hist2dplot(
            correlation,
            xbins=xbins,
            ybins=ybins,
            cbarextend=True,
            flow=None,
            ax=ax,
        )
        hist2d.cbar.set_label("Correlation coefficient")
        ax.set_xlabel(QUANTITIES[quantity]["label"])
        ax.set_ylabel(QUANTITIES[quantity]["label"])
        fig.set_size_inches(
            fig.get_size_inches()[0] * 1.05,
            fig.get_size_inches()[1],
        )
        fig.tight_layout()
        MatplotlibFormatter.dump(self.output(), fig)
