import matplotlib.pyplot as plt
import numpy as np
import uproot

import law
import luigi
import mplhep as hep
from analysis.framework.quantities import QUANTITIES, ybys_bins
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig
from analysis.framework.utils import DY_Sets
from analysis.tasks.PostProcessing import MergeUncertainties
from analysis.tasks.Theory import TheoryPrediction
from analysis.tools.plotting.axes_ticker import set_ticklabels
from analysis.tools.plotting.CustomLegend import custom_ybys_legend
from law.contrib.matplotlib import MatplotlibFormatter
from luigi.util import inherits


@inherits(ExcaliburConfig)
class StaggerPlot(AnalysisTask):
    dataset = luigi.Parameter(
        default="data"
        # default=DY_Sets.DYJetsToLL_combined.name
    )  # overwrite default from ExcaliburConfig
    input_is_mc = luigi.BoolParameter(default=False)
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    unfolding_weight = luigi.Parameter(default="totalWeight")
    gen_weight = luigi.Parameter("weight")

    order = luigi.Parameter(
        default="NNLO", description="Perturbative order of the theory calculation"
    )
    pdf = luigi.Parameter(
        default="PDF4LHC21_40",
        description="PDF to evaluate for theory predictions. Supported PDFs: all in `source scripts/ini_fastNLO.sh; lhapdf ls`",
    )

    selection_suffix = luigi.Parameter()

    def requires(self):
        req = {
            "unc": MergeUncertainties.req(self),
        }
        req["unfolded"] = req["unc"].requires()["unf"]
        req["theory"] = TheoryPrediction.req(self, order=self.order, pdf=self.pdf)
        return req

    def output(self):
        return law.LocalFileTarget(
            self.local_path(
                f"{self.excalibur_config_version}_{self.selection_suffix}_cuts",
                f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset}-{self.unfolding_weight}",
                "xs_overview.pdf",
            ),
            fs="local_plots",
        )

    def run(self):
        hep.style.use("CMS")
        fig, ax = plt.subplots()
        size = fig.get_size_inches()  # size in pixels
        fig.set_size_inches(size[0], size[1] * 1.2)
        hep.cms.label(label="Preliminary", data=True, lumi=138, loc=0, ax=ax)
        for _i, (ybys_bin, _d) in enumerate(ybys_bins.items()):
            theory, bins = uproot.open(self.input()["theory"].abspath)[
                f"{ybys_bin}/h_genzpt_weight"
            ].to_numpy()
            theory *= 10 ** _d["stagger"]  # stagger factor
            hep.histplot(
                theory,
                bins=bins,
                histtype="step",
                color="black",
                edges=False,
                label=(f"NNLO({self.pdf})\n" + r"$\times$NP" if _i == 0 else None),
            )
            data, bins = uproot.open(self.input()["unfolded"].abspath)[
                f"{ybys_bin}/h_zpt_{self.unfolding_weight}"
            ].to_numpy()
            unc_up, _ = uproot.open(self.input()["unc"].abspath)[
                f"{ybys_bin}/h_zpt_total_up"
            ].to_numpy()
            unc_dn, _ = uproot.open(self.input()["unc"].abspath)[
                f"{ybys_bin}/h_zpt_total_down"
            ].to_numpy()
            # I tested this, if one uses np.vstack, the down unceratinty has to be first
            unc = np.abs(np.vstack((unc_dn, unc_up)))
            data *= 10 ** _d["stagger"]
            unc *= 10 ** _d["stagger"]
            hep.histplot(
                data,
                bins=bins,
                xerr=0,
                yerr=unc,
                histtype="errorbar",
                color=_d["color"],
                marker=_d["marker"],
                fillstyle=_d["marker_style"],
            )
            if _i == 0:
                ax.errorbar(
                    x=[],
                    y=[],
                    xerr=True,
                    yerr=True,
                    color="black",
                    marker="D",
                    fillstyle="full",
                    label="Data",
                )
        ax.set_yscale("log")
        ax.set_xscale("log")
        ax.legend(loc="upper left")
        hep.yscale_legend(ax=ax)
        ybys_dicts = list(ybys_bins.values())
        for _d in ybys_dicts:
            _d["yb_mean"] = 0.5 * (_d["yb_range"][0] + _d["yb_range"][1])
            _d["ys_mean"] = 0.5 * (_d["ys_range"][0] + _d["ys_range"][1])
            _d["legend_bin_label"] = (
                r"$\mathdefault{{10^{{" + str(_d["stagger"]) + "}}}}$"
            )
        custom_ybys_legend(
            ax,
            ybys_dicts=ybys_dicts,
            borderpad=2.25,
            size_inches=2.75,
            loc="upper right",
            vertical_axis="yb",
            bin_label_key=("legend_bin_label"),
            x_tick_label_rotation=45,  # if args.figscale <= 0.7 else 0,  # heuristic
            shaded_bin_backgrounds=False,
            fontsize="x-small",
        )
        low, high = ax.get_ylim()
        ax.set_ylim(low, 10 ** (np.log10(high) * 1.15))
        ax.set_xlabel(QUANTITIES["zpt"]["label"])
        ax.set_ylabel(QUANTITIES["zpt"]["xs_label"])
        set_ticklabels(ax)
        fig.tight_layout()
        self.output().parent.touch()
        MatplotlibFormatter.dump(self.output(), fig)
