import matplotlib.pyplot as plt
import numpy as np
import uproot

import law
import luigi
import mplhep as hep
from analysis.framework.quantities import QUANTITIES, ybys_bins
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig
from analysis.framework.utils import DY_Sets
from analysis.tasks.PostProcessing import MergeUncertainties
from analysis.tasks.Theory import TheoryPrediction
from analysis.tools.plotting.axes_ticker import set_ticklabels
from analysis.tools.plotting.CustomLegend import custom_ybys_legend
from law.contrib.matplotlib import MatplotlibFormatter
from luigi.util import inherits

COLORS = ["red", "blue", "green"]
LINESTYLES = ["solid", "dotted", "dashed", "dashdot"]
HATCHES = ["O", "*", ".", "|"]


@inherits(ExcaliburConfig)
class TheoryDataComparison(AnalysisTask):
    dataset = luigi.Parameter(
        default="data"
        # default=DY_Sets.DYJetsToLL_combined.name
    )  # overwrite default from ExcaliburConfig
    input_is_mc = luigi.BoolParameter(default=False)
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    unfolding_weight = luigi.Parameter(default="totalWeight")
    gen_weight = luigi.Parameter("weight")

    order = luigi.Parameter(
        default="NNLO", description="Perturbative order of the theory calculation"
    )
    pdfs = luigi.ListParameter(
        default=[
            "PDF4LHC21_40",
            "NNPDF31_nnlo_as_0118",
            "CT18NNLO",
            "MSHT20nnlo_as118",
        ],
        description="PDFs to evaluate for theory predictions. Reference theory is for first PDF. Supported PDFs: all in `source scripts/ini_fastNLO.sh; lhapdf ls`",
    )

    selection_suffix = luigi.Parameter()

    def requires(self):
        req = {
            "unc": MergeUncertainties.req(self),
        }
        for pdf in self.pdfs:
            req[f"theory_{pdf}"] = TheoryPrediction.req(self, pdf=pdf)
        req["unfolded"] = req["unc"].requires()["unf"]
        return req

    def output(self):
        filelist = {}
        # individual bins
        for ybys_bin in ybys_bins:
            filelist[ybys_bin] = law.LocalFileTarget(
                self.local_path(
                    f"{self.excalibur_config_version}_{self.selection_suffix}_cuts",
                    f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset}-{self.unfolding_weight}",
                    f"{self.order}_data_ratio_{ybys_bin}.pdf",
                ),
                fs="local_plots",
            )
        # summary plot
        filelist["summary"] = law.LocalFileTarget(
            self.local_path(
                f"{self.excalibur_config_version}_{self.selection_suffix}_cuts",
                f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset}-{self.unfolding_weight}",
                f"{self.order}_data_ratio_summary.pdf",
            ),
            fs="local_plots",
        )
        return law.SiblingFileCollection(filelist)

    def run(self):
        self.output().dir.touch()
        hep.style.use("CMS")
        # summary plot
        fig, ax = plt.subplots()
        size = fig.get_size_inches()  # size in pixels
        fig.set_size_inches(size[0], size[1] * 1.4)
        hep.cms.label(label="Preliminary", data=True, lumi=138, loc=0, ax=ax)
        for _iybys, (ybys_bin, _d) in enumerate(ybys_bins.items()):
            # individual plots
            fig_single, ax_single = plt.subplots()
            hep.cms.label(label="Preliminary", data=True, lumi=138, loc=0, ax=ax_single)
            # compute ratios and plot
            theories = {}
            theory_uncertainties = {}
            for pdf in self.pdfs:
                theory, bins = uproot.open(self.input()[f"theory_{pdf}"].abspath)[
                    f"{ybys_bin}/h_genzpt_weight"
                ].to_numpy()
                theories[pdf] = theory
                theory_uncertainties[f"total_{pdf}"] = np.zeros(
                    shape=(np.vstack((theory, theory))).shape
                )
                thup, _ = uproot.open(self.input()[f"theory_{pdf}"].abspath)[
                    f"{ybys_bin}/h_genzpt_weight_total_up"
                ].to_numpy()
                thdn, _ = uproot.open(self.input()[f"theory_{pdf}"].abspath)[
                    f"{ybys_bin}/h_genzpt_weight_total_down"
                ].to_numpy()
                # use relative uncertainties
                theory_uncertainties[f"total_{pdf}"] = (
                    np.abs(np.vstack((thdn, thup))) / theories[pdf]
                )

            hep.histplot(
                np.ones(theories[self.pdfs[0]].shape) + _d["shift"],
                bins=bins,
                xerr=True,
                yerr=False,
                histtype="errorbar",
                color="black",
                marker=None,
                ax=ax,
            )
            hep.histplot(
                np.ones(theories[self.pdfs[0]].shape) + _d["shift"],
                bins=bins,
                yerr=theory_uncertainties["total_{pdf}".format(pdf=self.pdfs[0])],
                histtype="band",
                edges=False,
                edgecolor="black",
                hatch="oo",
                ax=ax,
            )
            hep.histplot(
                np.ones(theories[self.pdfs[0]].shape),
                bins=bins,
                xerr=True,
                yerr=False,
                histtype="errorbar",
                label="Theory",
                color="black",
                marker=None,
                ax=ax_single,
            )
            hep.histplot(
                np.ones(theories[self.pdfs[0]].shape),
                bins=bins,
                yerr=theory_uncertainties["total_{pdf}".format(pdf=self.pdfs[0])],
                histtype="band",
                label="Total theo. unc.",
                edges=False,
                edgecolor="black",
                hatch="oo",
                ax=ax_single,
            )
            for _ipdf, pdf in enumerate(self.pdfs[1:]):
                # hep.histplot(
                #     theories[pdf]/theories[self.pdfs[0]],
                #     bins=bins,
                #     yerr=theory_uncertainties[f"total_{pdf}"],
                #     histtype="band",
                #     edges=False,
                #     edgecolor=COLORS[_ipdf%len(COLORS)],
                #     facecolor="none",
                #     hatch=HATCHES[_ipdf%len(HATCHES)],
                #     ax=ax_single,
                # )
                ax_single.stairs(
                    values=theories[pdf] / theories[self.pdfs[0]],
                    edges=bins,
                    label=pdf,
                    color=COLORS[_ipdf % len(COLORS)],
                    linestyle=LINESTYLES[_ipdf % len(LINESTYLES)],
                    linewidth=4,
                    baseline=None,
                )
            data, bins = uproot.open(self.input()["unfolded"].abspath)[
                f"{ybys_bin}/h_zpt_{self.unfolding_weight}"
            ].to_numpy()
            data /= theory
            uncertainties = {}
            for name in ("stat", "total"):
                up, _ = uproot.open(self.input()["unc"].abspath)[
                    f"{ybys_bin}/h_zpt_{name}_up"
                ].to_numpy()
                dn, _ = uproot.open(self.input()["unc"].abspath)[
                    f"{ybys_bin}/h_zpt_{name}_down"
                ].to_numpy()
                uncertainties[name] = np.abs(np.vstack((dn, up))) / theory
            hep.histplot(
                data + _d["shift"],
                bins=bins,
                xerr=True,
                yerr=uncertainties["stat"],
                histtype="errorbar",
                color=_d["color"],
                marker=_d["marker"],
                fillstyle=_d["marker_style"],
                ax=ax,
            )
            hep.histplot(
                data + _d["shift"],
                bins=bins,
                yerr=uncertainties["total"],
                histtype="band",
                edges=False,
                edgecolor=_d["color"],
                hatch="\\\\\\\\  \\",
                ax=ax,
            )
            hep.histplot(
                data,
                bins=bins,
                xerr=True,
                yerr=uncertainties["stat"],
                histtype="errorbar",
                label="Data + stat. unc.",
                color=_d["color"],
                marker=_d["marker"],
                fillstyle=_d["marker_style"],
                ax=ax_single,
            )
            hep.histplot(
                data,
                bins=bins,
                yerr=uncertainties["total"],
                label="Total exp. unc.",
                histtype="band",
                edges=False,
                edgecolor=_d["color"],
                hatch="\\\\\\\\  \\",
                ax=ax_single,
            )
            ax_single.set_xscale("log")
            ax_single.set_xlim(QUANTITIES["zpt"]["range"])
            ax_single.set_ylim(0.8, 1.4)
            handles, labels = hep.sort_legend(
                ax_single,
                order=[
                    "Data + stat. unc.",
                    "Total exp. unc.",
                    "Theory",
                    "Total theo. unc.",
                ]
                + list(self.pdfs[1:]),
            )
            ax_single.text(
                0.03,
                0.95,
                "${} \\leq y_b < {}$\n".format(*_d["yb_range"])
                + "${} \\leq y^* < {}$".format(*_d["ys_range"]),
                fontsize="small",
                verticalalignment="top",
                fontproperties="Tex Gyre Heros",
                transform=ax_single.transAxes,
            )
            ax_single.legend(handles, labels, frameon=False)
            ax_single.set_xlabel(QUANTITIES["zpt"]["label"])
            ax_single.set_ylabel(
                r"Ratio to {order} ({pdf}) $\times$ NP".format(
                    order=self.order, pdf=self.pdfs[0]
                )
            )
            set_ticklabels(ax_single)
            fig_single.tight_layout()
            MatplotlibFormatter.dump(self.output().targets[ybys_bin], fig_single)
        # only for summary plot from now on
        ax.errorbar(
            x=[],
            y=[],
            xerr=True,
            yerr=True,
            color="black",
            marker="D",
            markersize=10.0,
            elinewidth=1,
            fillstyle="full",
            label="Data + stat. unc.",
        )
        ax.fill_between(
            x=[],
            y1=[],
            y2=[],
            edgecolor="black",
            facecolor="none",
            hatch="\\\\\\\\  \\",
            label="Total exp. unc.",
        )
        ax.plot([], [], color="black", marker=None, label="Theory")
        ax.fill_between(
            x=[],
            y1=[],
            y2=[],
            edgecolor="black",
            facecolor="none",
            hatch="oo",
            label="Total theo. unc.",
        )
        # axes, scales, style, and legend
        ax.set_xscale("log")
        handles, labels = hep.sort_legend(
            ax,
            order=[
                "Data + stat. unc.",
                "Total exp. unc.",
                "Theory",
                "Total theo. unc.",
            ],
        )
        ax.legend(handles, labels, loc="upper left", frameon=False)
        ybys_dicts = list(ybys_bins.values())
        for _d in ybys_dicts:
            _d["yb_mean"] = 0.5 * (_d["yb_range"][0] + _d["yb_range"][1])
            _d["ys_mean"] = 0.5 * (_d["ys_range"][0] + _d["ys_range"][1])
            if _d["shift"] > 0:
                _d["legend_bin_label"] = "$+{:.1f}$".format(_d["shift"])
            elif _d["shift"] == 0:
                _d["legend_bin_label"] = "$+0$"
            else:
                _d["legend_bin_label"] = "${:.1f}$".format(_d["shift"])
        custom_ybys_legend(
            ax,
            ybys_dicts=ybys_dicts,
            borderpad=2.25,
            size_inches=2.75,
            loc="upper right",
            vertical_axis="yb",
            bin_label_key=("legend_bin_label"),
            x_tick_label_rotation=45,  # if args.figscale <= 0.7 else 0,  # heuristic
            shaded_bin_backgrounds=False,
            fontsize="x-small",
        )
        low, high = ax.get_ylim()
        ax.set_ylim(0, 6.2)
        ax.set_xlabel(QUANTITIES["zpt"]["label"])
        ax.set_ylabel(
            r"Ratio to {order} ({pdf}) $\times$ NP $+X$".format(
                order=self.order, pdf=self.pdfs[0]
            )
        )
        set_ticklabels(ax)
        fig.tight_layout()
        MatplotlibFormatter.dump(self.output().targets["summary"], fig)
