import luigi
import os

from luigi.util import inherits
from analysis.tasks.PostProcessing.LumberjackRun import Histograms
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig
from analysis.framework.utils import run_command, karma_env


@inherits(ExcaliburConfig)
class CalcAcceptanceFakerate(AnalysisTask):
    selection_suffix = luigi.Parameter()
    gen_weight = luigi.Parameter(default="weight")
    reco_weights = luigi.ListParameter()

    def requires(self):
        return {
            "reco": Histograms.req(
                self,
                splitting="ybys",
                task="DATAShape",
                selections="reco{}".format(self.selection_suffix),
            ),
            "gen": Histograms.req(
                self,
                splitting="genybys",
                task="MCShape",
                selections="gen{}".format(self.selection_suffix),
            ),
            "response_gen": Histograms.req(
                self,
                splitting="genybys",
                task="CreateResponseMatrix",
                selections="response{}".format(self.selection_suffix),
            ),
            "response_reco": Histograms.req(
                self,
                splitting="ybys",
                task="CreateResponseMatrix",
                selections="response{}".format(self.selection_suffix),
            ),
        }

    def output(self):
        return self.local_target(
            self.excalibur_config_version,
            f"{self.selection_suffix}_cuts",
            self.dataset,
            "acceptance_fakerate.root",
        )

    def run(self):
        # create ouput folder for result
        try:
            os.makedirs(os.path.dirname(self.output().abspath))
        except OSError as e:
            if e.errno != 17:
                print("Error:", e)
        command = (
            [
                "palisade.py",
                "task",
                "zjet_excalibur",
                "rootfile_acceptance_fakerate",
                "--quantities",
                "zpt",
                "--response-gen-filename",
                self.input()["response_gen"].abspath,
                "--response-reco-filename",
                self.input()["response_reco"].abspath,
                "--reco-filename",
                self.input()["reco"].abspath,
                "--gen-filename",
                self.input()["gen"].abspath,
                "--gen-weight",
                self.gen_weight,
                "--reco-weights",
            ]
            + list(self.reco_weights)
            + ["--output-format-root", "{}".format(self.output().abspath)]
        )
        print(f"command: {command}")
        try:
            run_command(command, env=karma_env)
        except RuntimeError as e:
            self.output().remove()
            raise e


@inherits(ExcaliburConfig)
class FitAcceptanceFakerate(AnalysisTask):
    selection_suffix = luigi.Parameter()
    gen_weight = luigi.Parameter(default="weight")
    reco_weights = luigi.ListParameter()

    def requires(self):
        return {
            "reco": Histograms.req(
                self,
                splitting="ybys",
                task="DATAShape",
                selections="reco{}".format(self.selection_suffix),
            ),
            "gen": Histograms.req(
                self,
                splitting="genybys",
                task="MCShape",
                selections="gen{}".format(self.selection_suffix),
            ),
            "response_gen": Histograms.req(
                self,
                splitting="genybys",
                task="CreateResponseMatrix",
                selections="response{}".format(self.selection_suffix),
            ),
            "response_reco": Histograms.req(
                self,
                splitting="ybys",
                task="CreateResponseMatrix",
                excalibur_config_version=self.excalibur_config_version,
                selections="response{}".format(self.selection_suffix),
            ),
        }

    def output(self):
        return [
            self.local_plot_target(
                f"{self.excalibur_config_version}_{self.selection_suffix}_cuts",
                self.dataset,
            ),
            self.local_target(
                self.excalibur_config_version,
                f"{self.selection_suffix}_cuts",
                self.dataset,
                "acceptance_fakerate_fit.root",
            ),
        ]

    def run(self):
        # create ouput folder for result
        try:
            os.makedirs(os.path.dirname(self.output()[0].abspath))
        except OSError as e:
            if e.errno != 17:
                print("Error:", e)
        command = (
            [
                "palisade.py",
                "task",
                "zjet_excalibur",
                "fit_acceptance_fakerate",
                "--quantities",
                "zpt",
                "--data-period",
                self.data_period.replace("ul", ""),
                "--response-gen-filename",
                self.input()["response_gen"].abspath,
                "--response-reco-filename",
                self.input()["response_reco"].abspath,
                "--reco-filename",
                self.input()["reco"].abspath,
                "--gen-filename",
                self.input()["gen"].abspath,
                "--gen-weight",
                self.gen_weight,
                "--reco-weights",
            ]
            + list(self.reco_weights)
            + [
                "--output-format-pic",
                "{}/{{reco_weight}}/{{ybys[name]}}.pdf".format(
                    self.output()[0].abspath
                ),
                "--output-format-root",
                "{}".format(self.output()[1].abspath),
            ]
        )
        print(f"command: {command}")
        try:
            run_command(command, env=karma_env)
        except RuntimeError as e:
            self.output().remove()
            raise e


@inherits(ExcaliburConfig)
class UnravelAcceptanceFakerate(AnalysisTask):
    selection_suffix = luigi.Parameter()
    gen_weight = luigi.Parameter(default="weight")
    reco_weights = luigi.ListParameter()
    forward_smearing = luigi.BoolParameter(default=False)

    def requires(self):
        if self.forward_smearing:
            return FitAcceptanceFakerate.req(self)
        return CalcAcceptanceFakerate.req(self)

    def output(self):
        return self.local_target(
            self.excalibur_config_version,
            f"{self.selection_suffix}_cuts",
            self.dataset,
            "acceptance_fakerate{}.root".format("_FS" if self.forward_smearing else ""),
        )

    def run(self):
        if self.forward_smearing:
            input_file = self.input()[1].abspath
        else:
            input_file = self.input().abspath
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "unravel_acceptance_and_fakerate",
            "--filename-input",
            "{}".format(input_file),
            "--filename-output",
            "{}".format(self.output().abspath),
            "--reco-weights",
        ] + list(self.reco_weights)
        try:
            run_command(command, karma_env)
        except RuntimeError as e:
            os.remove(self.output().abspath)
            raise e
