import os
from abc import ABCMeta

import law
import luigi
from analysis.framework.dataset_map import cross_sections, dataset_map
from analysis.framework.tasks import AnalysisTask, LumberjackConfig
from analysis.framework.utils import (
    Energy_Shifts,
    data_periods_ul,
    jec_uncertainty_sources,
    karma_env,
    run_command,
)
from analysis.tasks.Excalibur import GetGeneratorWeights
from analysis.tasks.Excalibur.ExcaliburWorkflow import ExcaliburMerge
from analysis.tasks.Remote import HTCondorWorkflow
from law.decorator import localize, safe_output
from law.logger import get_logger
from luigi.util import inherits

logger = get_logger(__name__)

# current order of tasks is:
# Run -> Scale -> Merge/hadd -> Stitch -> HistosWrapper -> MergeShifts -> CombineRun2


@inherits(LumberjackConfig)
class LumberjackBaseWorkflow(law.LocalWorkflow, AnalysisTask):
    __metaclass__ = ABCMeta

    def create_branch_map(self):
        branches = {0: (Energy_Shifts.NONE, "Total")}
        if (
            "DYJetsToLL" in self.dataset
            and "M-10to50" not in self.dataset
            and "herwig" not in self.dataset
            and "MLM" not in self.dataset
            and "MiNNLO" not in self.dataset
        ):
            branches[1] = (Energy_Shifts.JER_UP, "Total")
            branches[2] = (Energy_Shifts.JER_DOWN, "Total")
            for i, jec_shift in enumerate(jec_uncertainty_sources, start=len(branches)):
                branches[i] = (Energy_Shifts.JEC_UP, jec_shift)
                branches[i + len(jec_uncertainty_sources)] = (
                    Energy_Shifts.JEC_DOWN,
                    jec_shift,
                )
        return branches

    def output(self):
        energy_shift, jec_source = self.branch_data
        target = self.local_target(
            self.excalibur_config_version,
            f"{self.dataset}",
            f"{self.task}_{self.splitting}_{self.selections}",
            "shift_{}.root".format("_".join([energy_shift.name, jec_source])),
        )
        return target


class LumberjackRun(HTCondorWorkflow, LumberjackBaseWorkflow):
    def output(self):
        energy_shift, jec_source = self.branch_data
        target = self.remote_target(
            self.excalibur_config_version,
            f"{self.dataset}",
            f"{self.task}_{self.splitting}_{self.selections}",
            "shift_{}.root".format("_".join([energy_shift.name, jec_source])),
        )
        return target

    def workflow_requires(self):
        return {
            branch: ExcaliburMerge.req(
                self, energy_shift=energy_shift, jec_uncertainty_source=jec_source
            )
            for branch, (energy_shift, jec_source) in self.branch_map.items()
        }

    def requires(self):
        return self.workflow_requires()[self.branch]

    @localize(input=True, output=False)
    def run(self):
        tmp_out = law.LocalDirectoryTarget(is_tmp=True)
        tmp_out.touch()
        input_type = "data" if "data" in self.dataset else "mc"

        tree = "{}zjetcuts_L1L2L3{}/ntuple".format(
            "gen" if "MC" in self.task else "", "Res" if "data" in self.dataset else ""
        )

        command = [
            "lumberjack.py",
            "-a",
            self.karma_config,
            "-i",
            self.input().abspath,
            "--selections",
            self.selections,
            "--tree",
            tree,
            "--overwrite",
            "-j",
            str(self.htcondor_cpus),
            "--input-type",
            input_type,
            "task",
            "{}_{}".format(self.task, self.splitting),
            "--output-dir",
            tmp_out.abspath,
            "--output-file-suffix",
            "{}".format(self.selections),
        ]
        run_command(command, karma_env)
        self.output().move_from_local(
            os.path.join(
                tmp_out.abspath, f"{self.task}_{self.splitting}_{self.selections}.root"
            )
        )


class LumberjackScale(LumberjackBaseWorkflow):
    def workflow_requires(self):
        if "data" in self.dataset:
            raise Exception("You want to scale data with lumi? Seriously?")
        return {"common": LumberjackRun.req(self, workflow="htcondor")}

    def requires(self):
        return self.workflow_requires()["common"].as_branch()

    def output(self):
        energy_shift, jec_source = self.branch_data
        target = self.remote_target(
            self.excalibur_config_version,
            f"{self.dataset}",
            f"{self.task}_{self.splitting}_{self.selections}",
            "shift_{}.root".format("_".join([energy_shift.name, jec_source])),
        )
        return target

    @safe_output
    def run(self):
        tmp_out = law.LocalFileTarget(is_tmp="root")
        tmp_out.parent.touch()
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "rootfile_scale_lumi",
            "--data-period",
            self.data_period.replace("ul", ""),
            "--lumberjack-task",
            self.task,
            "--filename",
            self.input().uri(),
            "--filename-output",
            tmp_out.abspath,
        ]
        run_command(command, karma_env)
        self.output().move_from_local(tmp_out)


class LumberjackHadd(LumberjackBaseWorkflow):
    def workflow_requires(self):
        req = {}
        # catch Run2 first
        if self.data_period == "Run2":
            raise NotImplementedError(
                "Run2 combination should be handled by the CombineRun2 Task, also when using hadd"
            )
        # merge 2016preVFP and postVFP into sinlge 2016 dataset
        if self.data_period == "2016ul":
            req["2016ulpreVFP"] = LumberjackHistos.req(self, data_period="2016ulpreVFP")
            req["2016ulpostVFP"] = LumberjackHistos.req(
                self, data_period="2016ulpostVFP"
            )
            return req
        # check data
        if self.dataset == "data":
            for dataset in dataset_map[self.data_period]:
                if self.dataset in dataset and "common" not in dataset:
                    req[dataset] = LumberjackHistos.req(self, dataset=dataset)
            return req
        # special case for 2016B
        if self.data_period == "2016ulpreVFP" and self.dataset == "dataB":
            for dataset in ("dataBver1", "dataBver2"):
                req[dataset] = LumberjackHistos.req(self, dataset=dataset)
            return req
        # DY Magic
        elif self.dataset == "DYJetsToLL_MiNNLO":
            for dataset in ("DYJetsToMuMu", "DYJetsToTauTau"):
                req[dataset] = LumberjackHistos.req(self, dataset=dataset)
            return req
        elif self.dataset == "DYJetsToLL_XJ":
            for dataset in ("DYJetsToLL_0J", "DYJetsToLL_1J", "DYJetsToLL_2J"):
                req[dataset] = LumberjackHistos.req(self, dataset=dataset)
            return req
        elif self.dataset == "DYJetsToLL_PtZ":
            for dataset in (
                "DYJetsToLL_PtZ_0To50",
                "DYJetsToLL_PtZ_50To100",
                "DYJetsToLL_PtZ_100To250",
                "DYJetsToLL_PtZ_250To400",
                "DYJetsToLL_PtZ_400To650",
                "DYJetsToLL_PtZ_650ToInf",
            ):
                req[dataset] = LumberjackHistos.req(self, dataset=dataset)
            return req
        elif self.dataset == "DYJetsToLL_PtZ_combined":
            for dataset in (
                "DYJetsToLL_PtZ_0To50",
                "DYJetsToLL_PtZ_50To100",
                "DYJetsToLL_PtZ_100To250",
                "DYJetsToLL_PtZ_250To400",
                "DYJetsToLL_PtZ_400To650",
                "DYJetsToLL_PtZ_650ToInf",
            ):
                req[dataset] = LumberjackHistos.req(
                    self, dataset=dataset, selections=self.selections
                )
            # work around buggy cut on gen level in central CMS production
            # https://cms-talk.web.cern.ch/t/bug-in-ul-pt-binned-dy-samples/11639
            # Cutting on gen_Pt > 0 essentially leads to removing all contribution with
            # more than 1 parton at LHE level, as there is nothing to recoil against.
            # So we need to get the 0 parton contribution with highes stat (0J samle).
            req["DYJetsToLL_0J"] = LumberjackHistos.req(
                self,
                dataset="DYJetsToLL_0J",
                selections=self.selections + "_0LHEPartons",
            )
            return req
        # SingleTop
        if self.dataset == "ST":
            for dataset in dataset_map[self.data_period]:
                if self.dataset in dataset and "common" not in dataset:
                    req[dataset] = LumberjackHistos.req(self, dataset=dataset)
            return req
        # WJets
        if self.dataset == "WJetsToLNu_XJ":
            for dataset in (
                "WJetsToLNu_0J",
                "WJetsToLNu_1J",
                "WJetsToLNu_2J",
            ):
                req[dataset] = LumberjackHistos.req(self, dataset=dataset)
            return req
        raise NotImplementedError

    def requires(self):
        req = [workflow.as_branch() for workflow in self.workflow_requires().values()]
        return req

    @safe_output
    def run(self):
        filelist = [x.uri() for x in self.input()]
        command = ["hadd", "-f", self.output().abspath] + filelist

        self.output().touch()
        run_command(command, env=karma_env)


class LumberjackStitch(LumberjackBaseWorkflow):
    dataset = luigi.Parameter(default="DYJetsToLL_combined")

    def workflow_requires(self):
        if self.dataset != "DYJetsToLL_combined":
            raise KeyError("LumberjackStitch is only defined for DYJetsToLL_combined!")
        req = {
            "inc": LumberjackHistos.req(self, dataset="DYJetsToLL"),
            "XJ": LumberjackHistos.req(self, dataset="DYJetsToLL_XJ"),
        }
        for dataset in (
            "DYJetsToLL",
            "DYJetsToLL_0J",
            "DYJetsToLL_1J",
            "DYJetsToLL_2J",
        ):
            req["gen_weights_" + dataset] = GetGeneratorWeights.req(
                self, dataset=dataset
            )
        return req

    def requires(self):
        req = {}
        for key, requirement in self.workflow_requires().items():
            if isinstance(requirement, law.BaseWorkflow):
                req[key] = requirement.as_branch()
            else:
                req[key] = requirement
        return req

    @safe_output
    def run(self):
        weight_inc = float(
            self.input()["gen_weights_DYJetsToLL"].load()["number_generated_events"]
        )
        weight_njet = 0
        for dataset in ("DYJetsToLL_0J", "DYJetsToLL_1J", "DYJetsToLL_2J"):
            n = float(
                self.input()["gen_weights_" + dataset].load()["number_generated_events"]
            )
            frac = (
                cross_sections[dataset]["cross_section"]
                / cross_sections["DYJetsToLL"]["cross_section"]
            )
            weight_njet += frac * n
        scale_inc = weight_inc / (weight_inc + weight_njet)
        scale_njet = weight_njet / (weight_inc + weight_njet)

        command = (
            [
                "palisade.py",
                "task",
                "zjet_excalibur",
                "rootfile_combineDY",
                "--lumberjack-task",
                self.task,
                "--filenames",
            ]
            + [self.input()[dataset].uri() for dataset in ("inc", "XJ")]
            + [
                "--scales",
                str(scale_inc),
                str(scale_njet),
                "--filename-output",
                self.output().abspath,
            ]
        )
        run_command(command, env=karma_env)


# This needs Wrapper Task in the first place! For correct MRO, so the output is the inputs!
class LumberjackHistos(law.task.base.WrapperTask, LumberjackBaseWorkflow):
    """This task is a intermediary task, providing the correct predecessor tasks for the different datasets."""

    def workflow_requires(self):
        if self.data_period == "Run2":
            raise KeyError(
                "LumberjackHistos should only be called for a single data period."
            )
        # Treat 2016 pre and postVFP as one dataset, fully correlated
        if self.data_period == "2016ul":
            return {"common": LumberjackHadd.req(self)}
        # Simply hadd data
        if self.dataset == "data":
            return {"common": LumberjackHadd.req(self)}
        # and don't scale single data eras!
        if "data" in self.dataset:
            return {"common": LumberjackRun.req(self, workflow="htcondor")}
        # Do DY merging
        if self.dataset == "DYJetsToLL_combined":
            return {"common": LumberjackStitch.req(self)}
        if self.dataset == "DYJetsToLL_XJ":
            return {"common": LumberjackHadd.req(self)}
        if self.dataset == "DYJetsToLL_PtZ":
            return {"common": LumberjackHadd.req(self)}
        if self.dataset == "DYJetsToLL_PtZ_combined":
            return {"common": LumberjackHadd.req(self)}
        if self.dataset == "DYJetsToLL_MiNNLO":
            return {"common": LumberjackHadd.req(self)}
        # SingelTop Merging
        if self.dataset == "ST":
            return {"common": LumberjackHadd.req(self)}
        # WJetsMerging
        if self.dataset == "WJetsToLNu_XJ":
            return {"common": LumberjackHadd.req(self)}
        # no scaling for BinCenters
        if self.task == "BinCenters":
            return {"common": LumberjackRun.req(self)}
        # Default to scaling MC with lumi
        return {"common": LumberjackScale.req(self)}

    def requires(self):
        return self.workflow_requires()["common"].as_branch()


@inherits(LumberjackConfig)
class MergeShifts(AnalysisTask):
    def requires(self):
        return LumberjackHistos.req(self)

    def output(self):
        return self.local_target(
            self.excalibur_config_version,
            f"{self.dataset}",
            f"{self.task}_{self.splitting}_{self.selections}.root",
        )

    @safe_output
    def run(self):
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "rootfile_merge_shifts",
            "--lumberjack-task",
            self.task,
            "--data-period",
            self.data_period,
            "--filename-output",
            self.output().abspath,
        ]
        for i, (energy_shift, jec_source) in self.requires().branch_map.items():
            if energy_shift == Energy_Shifts.NONE:
                arg = "--filename"
            elif energy_shift == Energy_Shifts.JER_UP:
                arg = "--jer-up"
            elif energy_shift == Energy_Shifts.JER_DOWN:
                arg = "--jer-down"
            elif energy_shift == Energy_Shifts.JEC_UP:
                arg = f"--jec-{jec_source}-up"
            elif energy_shift == Energy_Shifts.JEC_DOWN:
                arg = f"--jec-{jec_source}-down"
            else:
                raise KeyError(f"Unknown energy shift {energy_shift.name}")
            command.extend([arg, self.input()["collection"][i].uri()])
        run_command(command, karma_env)


@inherits(LumberjackConfig)
class LumberjackCombineRun2(AnalysisTask):
    def requires(self):
        req = {}
        if self.data_period != "Run2":
            raise KeyError("LumberjackCombineRun2 only wokrs for data period Run2!")
        for period in data_periods_ul:
            req[period] = Histograms.req(self, data_period=period)
        return req

    def output(self):
        dataset = str(self.dataset)
        return self.local_target(
            self.excalibur_config_version,
            f"{dataset}",
            f"{self.task}_{self.splitting}_{self.selections}.root",
            name=LumberjackRun.__name__,
        )

    @safe_output
    def run(self):
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "rootfile_mergeRun2",
            "--lumberjack-task",
            self.task,
            "--filename-output",
            self.output().abspath,
            "--filenames",
        ]
        if (
            self.task == "MCShape"
            or self.task == "DATAShape_data"
            or self.task == "BinCenters"
            or "DY" not in self.dataset
            or "M-10to50" in self.dataset
            or "herwig" in self.dataset
        ):
            command = ["hadd", self.output().abspath]
        command.extend([self.input()[period].uri() for period in data_periods_ul])

        self.output().makedirs()
        run_command(command, env=karma_env)


@inherits(LumberjackConfig)
class Histograms(law.task.base.WrapperTask):
    """This Wrapper defines the files containing the final histograms for further processing."""

    def requires(self):
        task = self.task
        if task == "DATAShape":
            task = "{}_{}".format(task, "data" if "data" in str(self.dataset) else "mc")
        # handle Run2 combination methods
        if self.data_period == "Run2":
            return LumberjackCombineRun2.req(self, task=task)
        # include Energy Shifts for further processing in the same file
        if (
            "DYJetsToLL" in self.dataset
            and "M-10to50" not in self.dataset
            and "herwig" not in self.dataset
            and "MLM" not in self.dataset
            and "MiNNLO" not in self.dataset
            and task != "MCShape"
            and task != "BinCenters"
        ):
            return MergeShifts.req(self, task=task)
        return LumberjackHistos.req(self, task=task)

    def output(self):
        if isinstance(self.requires(), law.BaseWorkflow):
            return self.input()["collection"][0]
        return super().output()
