import luigi
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig
from analysis.framework.utils import DY_Sets, karma_env, run_command
from analysis.tasks.PostProcessing.LumberjackRun import Histograms
from analysis.tasks.PostProcessing.Unfold import Unfold
from law.decorator import safe_output
from luigi.util import inherits


@inherits(ExcaliburConfig)
class Reravel(AnalysisTask):
    dataset = luigi.Parameter(default=DY_Sets.DYJetsToLL_combined.name)
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    reco_weights = luigi.ListParameter()
    input_is_mc = luigi.BoolParameter(default=False)

    bkg_variation_sfs = luigi.ListParameter(default=None)

    forward_smearing = luigi.BoolParameter(default=False)
    selection_suffix = luigi.Parameter()

    def requires(self):
        reco_weights = list(self.reco_weights)
        try:
            reco_weights.remove("totalWeight_bkgsf_down")
        except ValueError:
            print("Background down variation not required!")
            pass
        try:
            reco_weights.remove("totalWeight_bkgsf_up")
        except ValueError:
            print("Background up variation not required!")
            pass
        return {
            "Data": Histograms(
                splitting="ybys",
                data_period=self.data_period,
                dataset=self.dataset,
                selections="reco{}".format(self.selection_suffix),
                task="DATAShape",
                excalibur_config_version=self.excalibur_config_version,
            ),
            "Unfold": Unfold.req(self, reco_weights=reco_weights),
        }

    def output(self):
        forward_smearing = "_FS" if self.forward_smearing else ""
        return self.local_target(
            self.excalibur_config_version,
            self.selection_suffix + "_cuts",
            f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset.name}{forward_smearing}.root",
        )

    @safe_output
    def run(self):
        command = (
            [
                "palisade.py",
                "task",
                "zjet_excalibur",
                "rootfile_reravel_cross_section",
                "--quantities",
                "zpt",
                "--weights",
            ]
            + list(self.reco_weights)
            + [
                "--data-period",
                self.data_period.replace("ul", ""),
                "--filename-unraveled",
                self.input()["Unfold"].abspath,
                "--filename-model",
                self.input()["Data"].abspath,
                "--output-file",
                self.output().abspath,
            ]
        )
        run_command(command, karma_env)


@inherits(ExcaliburConfig)
class UnravelCrossSection(AnalysisTask):
    dataset = luigi.Parameter(default="data")  # overwrite default from ExcaliburConfig
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    reco_weights = luigi.ListParameter()
    input_is_mc = luigi.BoolParameter(default=False)

    bkg_variation_sfs = luigi.ListParameter(default=[1.5, 0.5])

    forward_smearing = luigi.BoolParameter(default=False)
    selection_suffix = luigi.Parameter()

    def requires(self):
        return Reravel.req(self)

    def output(self):
        forward_smearing = "_FS" if self.forward_smearing else ""
        return self.local_target(
            self.excalibur_config_version,
            self.selection_suffix + "_cuts",
            f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset.name}{forward_smearing}.root",
        )

    @safe_output
    def run(self):
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "unravel_uncertainties",
            "--filename-input",
            self.input().abspath,
            "--filename-output",
            self.output().abspath,
            "--weights",
            "totalWeight",
        ]
        run_command(command, karma_env)
