import os

import law
import luigi
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig, LumberjackConfig
from analysis.framework.utils import (
    DY_Sets,
    bg_names,
    data_periods_ul,
    jec_uncertainty_sources,
    karma_env,
    run_command,
)
from analysis.tasks.PostProcessing.LumberjackRun import Histograms
from analysis.tasks.PostProcessing.Reravel import Reravel
from law.decorator import safe_output
from luigi.util import inherits


@inherits(LumberjackConfig)
class MergeMCShapeUncertainties(AnalysisTask):
    quantities = luigi.ListParameter()
    bkg_variation_sfs = luigi.ListParameter(default=[1.5, 0.5])

    def requires(self):
        req = dict()
        req["DY"] = Histograms.req(self)
        for dataset in bg_names:
            req[dataset] = Histograms.req(self, dataset=dataset)
        return req

    def output(self):
        task = self.task
        if "DATAShape" in task:
            task = "{}_{}".format(task, "data" if "data" in self.dataset else "mc")
        return self.local_target(
            self.excalibur_config_version,
            f"{self.dataset}",
            f"{task}_{self.splitting}_{self.selections}.root",
        )

    @safe_output
    def run(self):
        try:
            os.makedirs(os.path.dirname(self.output().abspath))
        except OSError as e:
            if e.errno != 17:
                raise e  # other than file exist error
        # build palisade command
        command = (
            [
                "palisade.py",
                "task",
                "zjet_excalibur",
                "rootfile_uncertainties_mc_shapes",
                "-q",
            ]
            + list(self.quantities)
            + [
                "--data-period",
                self.data_period.replace("ul", ""),
                "--filename-main",
                self.input()["DY"].abspath,
                "--bkg-variations",
            ]
            + list(map(str, self.bkg_variation_sfs))
            + ["--filenames-bkg"]
            + [str(self.input()[bkg].abspath) for bkg in bg_names]
            + [
                "--filename-output",
                self.output().abspath,
            ]
        )
        # run palisade
        run_command(command, env=karma_env)


@inherits(ExcaliburConfig)
class MergeUncertainties(AnalysisTask):
    dataset = luigi.Parameter(
        # default="data"
        default=DY_Sets.DYJetsToLL_combined.name
    )  # overwrite default from ExcaliburConfig
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    mc_dataset_alt = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_MLM)
    input_weight = luigi.Parameter(default="weight")
    input_is_mc = luigi.BoolParameter(default=False)

    bkg_variation_sfs = luigi.ListParameter(default=[1.5, 0.5])

    forward_smearing = luigi.BoolParameter(default=False)
    selection_suffix = luigi.Parameter()

    def requires(self):
        req = dict()
        reco_weights = ["totalWeight"]
        shifts = [
            "PileUp",
            "prefiring",
            "puJetID",
            "reco",
            "ID",
            "Iso",
            "trigger",
            "bkgsf",
            "lumi_2016",
            "lumi_2017",
            "lumi_2018",
            "lumi_13TeV_correlated",
            "lumi_13TeV_1718",
        ]
        if self.data_period == "Run2":
            shifts += [
                "jer_{}".format(era.replace("ul", "")) for era in data_periods_ul
            ]
            for shift in jec_uncertainty_sources:
                shifts += [
                    "jec_{}_{}".format(shift, era.replace("ul", ""))
                    for era in data_periods_ul
                ]
        else:
            shifts += ["jer"]
            shifts += [f"jec_{source}" for source in jec_uncertainty_sources]

        for shift in shifts:
            reco_weights += [f"totalWeight_{shift}_up", f"totalWeight_{shift}_down"]
        req["unf"] = Reravel.req(
            self,
            reco_weights=reco_weights,
        )
        req["unf_smooth"] = UnfoldSmooth.req(
            self,
            reco_weights=reco_weights,
        )
        req["unf_syst"] = Reravel.req(
            self,
            mc_dataset=self.mc_dataset_alt,
            reco_weights=["totalWeight"],
            bkg_variation_sfs=None,
        )
        return req

    def output(self):
        forward_smearing = "_FS" if self.forward_smearing else ""
        return self.local_target(
            self.excalibur_config_version,
            self.selection_suffix + "_cuts",
            f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset.name}{forward_smearing}.root",
        )

    @safe_output
    def run(self):
        self.output().parent.touch()
        # build palisade command
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "rootfile_uncertainties",
            "-q",
            "zpt",
            "--data-period",
            self.data_period.replace("ul", ""),
            "--filename-main",
            self.input()["unf"].abspath,
            "--filename-smooth",
            self.input()["unf_smooth"].abspath,
            "--filename-unf-alt",
            self.input()["unf_syst"].abspath,
            "--filename-output",
            self.output().abspath,
        ]
        # run palisade
        run_command(command, env=karma_env)


@inherits(ExcaliburConfig)
class UnravelUncertainties(AnalysisTask):
    """Unravelled uncertainties, only used when plotting by bin-index"""

    dataset = luigi.Parameter(default="data")  # overwrite default from ExcaliburConfig
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    input_is_mc = luigi.BoolParameter(default=False)

    forward_smearing = luigi.BoolParameter(default=False)
    selection_suffix = luigi.Parameter()

    unc_weights = luigi.ListParameter()

    def requires(self):
        return MergeUncertainties.req(self)

    def output(self):
        forward_smearing = "_FS" if self.forward_smearing else ""
        return self.local_target(
            self.excalibur_config_version,
            self.selection_suffix + "_cuts",
            f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset.name}{forward_smearing}.root",
        )

    @safe_output
    def run(self):
        self.output().parent.touch()
        # build palisade command
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "unravel_uncertainties",
            "--filename-input",
            self.input().abspath,
            "--filename-output",
            self.output().abspath,
            "--weights",
        ] + list(self.unc_weights)
        # run palisade
        run_command(command, env=karma_env)


@inherits(Reravel)
class UnfoldSmooth(AnalysisTask):
    def requires(self):
        return {
            "unf": Reravel.req(self),
            "bin_centers": Histograms.req(
                self,
                selections="reco{}".format(self.selection_suffix),
                task="BinCenters",
                splitting="ybys",
            ),
        }

    def output(self):
        forward_smearing = "_FS" if self.forward_smearing else ""
        return self.local_target(
            self.excalibur_config_version,
            self.selection_suffix + "_cuts",
            f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset.name}{forward_smearing}.root",
        )

    def run(self):
        tmp_file = law.LocalFileTarget(is_tmp="root")
        # build palisade command
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "rootfile_smooth_cross_section",
            "-q",
            "zpt",
            "--filename-input",
            self.input()["unf"].abspath,
            "--filename-bincenters",
            self.input()["bin_centers"].abspath,
            "--input-weight",
            self.input_weight,
            "--filename-output",
            tmp_file.abspath,
        ]
        command += ["--reco-weights"] + list(self.reco_weights)
        # run palisade
        run_command(command, env=karma_env)
        tmp_file.move_to(self.output())
