import law
import luigi
from analysis.framework.tasks import AnalysisTask, ExcaliburConfig
from analysis.framework.utils import DY_Sets, bg_names, karma_env, run_command
from analysis.tasks.PostProcessing.AccepatanceFakerate import UnravelAcceptanceFakerate
from analysis.tasks.PostProcessing.LumberjackRun import Histograms
from law.decorator import safe_output
from luigi.util import inherits


@inherits(ExcaliburConfig)
class UnfoldWorkflow(law.LocalWorkflow, AnalysisTask):
    dataset = luigi.Parameter(default=DY_Sets.DYJetsToLL_combined.name)
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    reco_weights = luigi.ListParameter()
    bkg_variation_sfs = luigi.ListParameter(default=None)
    forward_smearing = luigi.BoolParameter(default=False)
    input_is_mc = luigi.BoolParameter(default=False)
    selection_suffix = luigi.Parameter()

    def create_branch_map(self):
        return self.reco_weights

    def workflow_requires(self):
        req = {
            "input": Histograms(
                data_period=self.data_period,
                dataset=self.dataset,
                excalibur_config_version=self.excalibur_config_version,
                selections="reco{}".format(self.selection_suffix),
                task="DATAShape",
                splitting="ybys",
            ),
        }

        if self.forward_smearing:
            raise NotImplementedError("Forward Smearing not yet implemented!")

        req["response"] = Histograms(
            data_period=self.data_period,
            dataset=self.mc_dataset.name,
            excalibur_config_version=self.excalibur_config_version,
            selections="response{}".format(self.selection_suffix),
            task="CreateResponseMatrix",
            splitting="genybys",
        )
        req["acceptance_fakerate"] = UnravelAcceptanceFakerate(
            data_period=self.data_period,
            dataset=self.mc_dataset.name,
            excalibur_config_version=self.excalibur_config_version,
            selection_suffix=self.selection_suffix,
            reco_weights=self.reco_weights,
        )

        for bg in bg_names:
            req[bg] = Histograms(
                data_period=self.data_period,
                dataset=bg,
                excalibur_config_version=self.excalibur_config_version,
                selections="reco{}".format(self.selection_suffix),
                task="DATAShape",
                splitting="ybys",
            )

        return req

    def requires(self):
        return self.workflow_requires()

    def output(self):
        forward_smearing = "_FS" if self.forward_smearing else ""
        return self.local_target(
            self.excalibur_config_version,
            self.selection_suffix + "_cuts",
            f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset.name}{forward_smearing}",
            f"{self.branch_data}.root",
        )

    def run(self):
        tmp_file = law.LocalFileTarget(is_tmp="root")
        # build palisade command
        command = [
            "palisade.py",
            "task",
            "zjet_excalibur",
            "rootfile_unravel_unfold",
            "-q",
            "zpt",
            "--filename-input",
            self.input()["input"].abspath,
            "--filename-response",
            self.input()["response"].abspath,
            "--filename-acceptance-fakerate",
            self.input()["acceptance_fakerate"].abspath,
            "--input-weight-data",
            self.input_weight,
            "--nominal-weight-mc",
            self.reco_weights[0],
            "--reco-weights",
            self.branch_data,
            "--filename-output",
            tmp_file.abspath,
        ]
        command += ["--filenames-background"]
        for bg in bg_names:
            command.append(str(self.input()[bg].abspath))
        if not self.input_is_mc:
            command += ["--input-is-data"]
        if self.bkg_variation_sfs and self.branch == 0:
            command += ["--bkg-variations"] + list(map(str, self.bkg_variation_sfs))
        # run palisade
        run_command(command, env=karma_env)
        tmp_file.move_to(self.output())


@inherits(ExcaliburConfig)
class Unfold(AnalysisTask):
    dataset = luigi.Parameter(default=DY_Sets.DYJetsToLL_combined.name)
    mc_dataset = luigi.EnumParameter(enum=DY_Sets, default=DY_Sets.DYJetsToLL_combined)
    input_weight = luigi.Parameter(default="weight")
    reco_weights = luigi.ListParameter()
    bkg_variation_sfs = luigi.ListParameter(default=None)
    forward_smearing = luigi.BoolParameter(default=False)
    input_is_mc = luigi.BoolParameter(default=False)
    selection_suffix = luigi.Parameter()

    def requires(self):
        return UnfoldWorkflow.req(self)

    def output(self):
        forward_smearing = "_FS" if self.forward_smearing else ""
        return self.local_target(
            self.excalibur_config_version,
            self.selection_suffix + "_cuts",
            f"{self.dataset}-{self.input_weight}_unfolded_with_{self.mc_dataset.name}{forward_smearing}.root",
        )

    @safe_output
    def run(self):
        filelist = self.input()["collection"].uri()
        command = ["hadd", self.output().abspath] + filelist

        self.output().parent.touch()
        run_command(command, env=karma_env)
