import os

import law
import luigi

law.contrib.load("tasks", "cms", "wlcg", "git")


class BundleRepo(law.git.BundleGitRepository, law.tasks.TransferLocalFile):
    replicas = luigi.IntParameter(
        default=10,
        description="number of replicas to generate; default: 10",
    )

    exclude_files = ["tmp", "*~", "*.pyc", ".vscode/"]

    version = None
    task_namespace = None
    default_store = "$ANALYSIS_BASE/tmp/bundles"

    def get_repo_path(self):
        # required by BundleGitRepository
        return os.environ["ANALYSIS_BASE"]

    def single_output(self):
        repo_base = os.path.basename(self.get_repo_path())
        path = "{}.{}.tgz".format(os.path.basename(repo_base), self.checksum)
        return law.wlcg.WLCGFileTarget(os.path.join("bundles", path), fs="eos_store")

    def get_file_pattern(self):
        path = os.path.expandvars(os.path.expanduser(self.single_output().abspath))
        return self.get_replicated_path(path, i=None if self.replicas <= 0 else "*")

    def output(self):
        return law.tasks.TransferLocalFile.output(self)

    @law.decorator.safe_output
    def run(self):
        # create the bundle
        bundle = law.LocalFileTarget(is_tmp="tgz")
        self.bundle(bundle)

        # log the size
        self.publish_message(
            "bundled repository archive, size is {}".format(
                law.util.human_bytes(bundle.stat().st_size, fmt=True),
            )
        )

        # transfer the bundle
        self.transfer(bundle)


class BundleCMSSW(law.cms.BundleCMSSW, law.tasks.TransferLocalFile):
    replicas = luigi.IntParameter(
        default=10,
        description="number of replicas to generate; default: 10",
    )

    version = None
    task_namespace = None
    exclude = "^src/tmp"
    default_store = "$ANALYSIS_BASE/tmp/bundles"

    def get_cmssw_path(self):
        return os.environ["CMSSW_DIR"]

    def single_output(self):
        path = "{}.{}.tgz".format(
            os.path.basename(self.get_cmssw_path()), self.checksum
        )
        return law.wlcg.WLCGFileTarget(os.path.join("bundles", path), fs="eos_store")

    def get_file_pattern(self):
        path = os.path.expandvars(os.path.expanduser(self.single_output().abspath))
        return self.get_replicated_path(path, i=None if self.replicas <= 0 else "*")

    def output(self):
        return law.tasks.TransferLocalFile.output(self)

    def run(self):
        # create the bundle
        bundle = law.LocalFileTarget(is_tmp="tgz")
        self.bundle(bundle)

        # log the size
        self.publish_message(
            "bundled CMSSW archive, size is {}".format(
                law.util.human_bytes(bundle.stat().st_size, fmt=True),
            )
        )

        # transfer the bundle
        self.transfer(bundle)
