#!/usr/bin/env bash

# Bootstrap file that is executed by remote jobs submitted by law to set up the environment. This
# file contains multiple different bootstrap functions of which only one is invoked by the last line
# of this file. So-called render variables, denoted by "{{name}}", are replaced with variables
# configured in the remote workflow tasks, e.g. in HTCondorWorkflow.htcondor_job_config(), upon job

# Bootstrap function for standalone htcondor jobs, i.e., each jobs fetches a software and repository
# code bundle and unpacks them to have a standalone environment, independent of the submitting one.
# The setup script of the repository is sourced with a few environment variables being set before,
# tailored for remote jobs.
bootstrap_htcondor_standalone() {
    # set env variables
    local BASE_DIR="${PWD}"
    export HOME="${PWD}"
    export LCG_DIR="/cvmfs/grid.cern.ch/centos7-ui-200122"
    export ANALYSIS_BASE="${BASE_DIR}/repo"
    export CMSSW_BASE="${BASE_DIR}/cmssw"
    export CMSSW_VERSION="CMSSW_10_6_34"
    export SCRAM_ARCH="slc7_amd64_gcc700"

    # source the law wlcg tools, mainly for law_wlcg_get_file
    source "{{wlcg_tools}}" "" || return "$?"

    # when gfal-* executables are not available, source the lcg dir
    if ! law_wlcg_check_executable "gfal-ls" "silent" && [ -n "${LCG_DIR}" ] && [ -d "${LCG_DIR}" ]; then
        echo "gfal-* executables not available, sourcing from ${LCG_DIR}/etc/profile.d/setup-c7-ui-python3-example.sh"
        source "${LCG_DIR}/etc/profile.d/setup-c7-ui-python3-example.sh" || return "$?"
    fi

    # load the repo bundle
    (
        mkdir -p "${ANALYSIS_BASE}"
        cd "${ANALYSIS_BASE}" || exit "$?"
        law_wlcg_get_file "{{repo_uris}}" '{{repo_pattern}}' "${ANALYSIS_BASE}/repo.tgz" || return "$?"
        if [ ! -f "repo.tgz" ]; then
            echo "Repo bundle does not exist. Exiting..."
            return 1
        fi
        tar -xzf "repo.tgz" || return "$?"
        rm -f "repo.tgz"
        echo "Analysis repo setup done."
    )

    # load cmssw
    (
        source /cvmfs/cms.cern.ch/cmsset_default.sh
        mkdir -p "${CMSSW_BASE}"
        cd "${CMSSW_BASE}" || return "$?"
        scramv1 project CMSSW "${CMSSW_VERSION}"
        cd "${CMSSW_VERSION}" || return "$?"
        law_wlcg_get_file "{{cmssw_uris}}" '{{cmssw_pattern}}' "${PWD}/cmssw.tgz" || return "$?"
        if [ ! -f "cmssw.tgz" ]; then
            echo "CMSSW bundle does not exist. Exiting..."
            return 1
        fi
        tar -xzf "cmssw.tgz" || return "$?"
        rm -f "cmssw.tgz"
        cd "src" || return "$?"
        eval "$( scramv1 runtime -sh )"
        scram b
        echo "CMSSW setup done."
    )

    # source the repo setup
    echo "export CMSSW_DIR=${CMSSW_BASE}/${CMSSW_VERSION}" > "$ANALYSIS_BASE/cmssw.conf"
    cd "${ANALYSIS_BASE}" || return "$?"
    sleep 10  # filesystem catch-up
    source "${ANALYSIS_BASE}/setup.sh" || return "$?"

    return "0"
}

export ZJET_REMOTE=1
bootstrap_htcondor_standalone "$@"
