import json
import os

import law
import luigi
from analysis.framework.dataset_map import dataset_map
from analysis.framework.utils import kappa_env, run_command
from analysis.tasks.Remote import HTCondorWorkflow


class GetFilelist(law.Task):
    data_period = luigi.Parameter()
    dataset = luigi.Parameter()

    def output(self):
        return law.LocalFileTarget(
            f"filelist/{self.data_period}/{self.dataset}.json", fs="analysis_data"
        )

    def run(self):
        das_query = "file dataset={}".format(
            dataset_map[self.data_period][self.dataset]["dataset"]
        )
        code, out, error = run_command(
            [
                "/cvmfs/cms.cern.ch/common/dasgoclient",
                "--query",
                das_query,
                "--json",
            ],
            env=kappa_env,
        )
        out_json = json.loads(out)
        self.output().dump(out_json, indent=4)


class SkimDataset(HTCondorWorkflow, law.LocalWorkflow):
    data_period = luigi.Parameter()
    dataset = luigi.Parameter()
    xrootd_redirector = luigi.Parameter(
        default="root://xrootd-cms.infn.it/", significant=False
    )
    exclude_params_repr = {"xrootd_redirector"}

    output_collection_cls = law.NestedSiblingFileCollection

    def workflow_requires(self):
        return {"filelist": GetFilelist.req(self)}

    def create_branch_map(self):
        filelist = self.workflow_requires()["filelist"]
        if not filelist.output().exists():
            filelist.run()
        json = filelist.output().load()
        return {i: _d["file"][0] for i, _d in enumerate(json)}

    def output(self):
        return law.wlcg.WLCGFileTarget(
            f"Skimming{self.branch_data['name']}", fs="nrg_base"
        )

    def run(self):
        out_file = law.LocalFileTarget(is_tmp="root")
        kappa_config = {
            "2016ulpreVFP": "zjet16ul/kappaSkim_default_cfg_preVFP.py",
            "2016ulpostVFP": "zjet16ul/kappaSkim_default_cfg_postVFP.py",
            "2017ul": "zjet17ul/kappaSkim_default_cfg.py",
            "2018ul": "zjet18ul/kappaSkim_default_cfg.py",
        }

        inputFile = f"{self.xrootd_redirector}{self.branch_data['name']}"
        local_copy = law.LocalFileTarget(
            f"broken_files_from_tape{self.branch_data['name']}", fs="local_ceph"
        )
        if local_copy.exists():
            inputFile = local_copy.uri()

        isData = "data" in self.dataset
        global_tag = (
            dataset_map[self.data_period]["common_data"]["global_tag"]
            if isData
            else dataset_map[self.data_period]["common_mc"]["global_tag"]
        )
        kappa_py_cfg = kappa_config[self.data_period]
        if "herwig" in self.dataset:
            kappa_py_cfg = kappa_py_cfg.replace("default", "herwig")
        run_command(
            [
                "cmsRun",
                kappa_py_cfg,
                f"isData={isData}",
                f"globalTag={global_tag}",
                f"inputFiles={inputFile}",
                f"outputFile={out_file.uri()}",
            ],
            cwd=os.path.expandvars("$CMSSW_DIR/src/Kappa/Skimming/zjet/configs"),
            env=kappa_env,
        )

        self.output().move_from_local(out_file)


class SkimAll(law.WrapperTask):
    data_period = luigi.Parameter()

    def requires(self):
        return {
            dataset: SkimDataset.req(self, dataset=dataset)
            for dataset in dataset_map[self.data_period].keys()
            if "common" not in dataset
        }
