import io
import os

import numpy as np
import ROOT

import law
import luigi
from analysis.framework.quantities import ybys_bins
from analysis.framework.utils import run_command, set_environment_variables
from law.decorator import safe_output


class Grid(law.ExternalTask):
    order = luigi.Parameter(default="NNLO")

    def output(self):
        targets = {}
        for ybys_bin, _d in ybys_bins.items():
            yb, _ = _d["yb_range"]
            ys, _ = _d["ys_range"]
            yb = int(2 * yb)
            ys = int(2 * ys)
            targets[ybys_bin] = law.LocalFileTarget(
                f"grids/zj5362zpt-v2/Combined/Final/ZJ.{self.order}.zj5362zpt-v2_yb{yb}_ys{ys}_ptz_v26.tab.gz",
                fs="local_ceph",
            )
        return law.SiblingFileCollection(targets)


class GridEval(law.Task):
    order = luigi.Parameter(default="NNLO")
    pdf = luigi.Parameter(default="PDF4LHC21_40")

    def requires(self):
        return Grid.req(self)

    def output(self):
        return law.LocalFileTarget(
            f"Theory/{self.order}_{self.pdf}.root",
            fs="local_store",
        )

    @safe_output  # automatically delete ouptut file if task fails
    def run(self):
        self.output().parent.touch()
        fastnlo_env = set_environment_variables(
            os.path.expandvars("$ANALYSIS_BASE/scripts/ini_fastNLO.sh")
        )
        out_file = ROOT.TFile.Open(self.output().abspath, "RECREATE")
        for ybys_bin, _d in ybys_bins.items():
            out_file.mkdir(ybys_bin)
            out_file.cd(ybys_bin)
            for unc, fnlo_unc in {"scale": "6P", "pdf": "L6", "stat": "ST"}.items():
                code, out, error = run_command(
                    [
                        "fnlo-tk-yodaout",
                        self.input()[ybys_bin].abspath,
                        self.pdf,
                        fnlo_unc,
                        self.order,
                    ],
                    fastnlo_env,
                )
                f_handler = io.StringIO(out)
                idx, mean, down, up = np.loadtxt(
                    f_handler, comments=["#"], skiprows=18, unpack=True
                )
                down *= mean  # make absolute
                up *= mean  # make absolute
                assert len(_d["zpt_binning"]) == len(idx) + 1
                hist_mean = ROOT.TH1D(
                    "h_genzpt_weight",
                    f"{ybys_bin}_genzpt",
                    len(idx),
                    _d["zpt_binning"],
                )
                hist_down = ROOT.TH1D(
                    f"h_genzpt_weight_{unc}_down",
                    f"{ybys_bin}_genzpt_{unc}_down",
                    len(idx),
                    _d["zpt_binning"],
                )
                hist_up = ROOT.TH1D(
                    f"h_genzpt_weight_{unc}_up",
                    f"{ybys_bin}_genzpt_{unc}_up",
                    len(idx),
                    _d["zpt_binning"],
                )
                for i, m, d, u in zip(idx, mean, down, up):
                    i = int(i)
                    hist_mean.SetBinContent(i, m)
                    hist_down.SetBinContent(i, d)
                    hist_up.SetBinContent(i, u)
                hist_mean.Write()
                hist_down.Write()
                hist_up.Write()
                del hist_mean
                del hist_down
                del hist_up
        out_file.Close()
