import base64
import enum
import json

import matplotlib.pyplot as plt
import numpy as np
import ROOT

import law
import luigi
import mplhep as hep
from analysis.framework.quantities import QUANTITIES, ybys_bins
from analysis.tools.plotting.axes_ticker import set_ticklabels
from analysis.tools.plotting.legend_helpers import merge_legend_handles_labels
from analysis.tools.plotting.styles import colormap_5_no_grey
from law.contrib.matplotlib import MatplotlibFormatter
from law.decorator import safe_output

__all__ = ["NPCalc", "NPCorrections"]


def json_numpy_obj_hook(dct):
    """
    Decodes a previously encoded numpy ndarray
    with proper shape and dtype
    :param dct: (dict) json encoded ndarray
    :return: (ndarray) if input was an encoded ndarray
    """
    if isinstance(dct, dict) and "__ndarray__" in dct:
        data = base64.b64decode(dct["__ndarray__"])
        return np.frombuffer(data, dct["dtype"]).reshape(dct["shape"])
    elif isinstance(dct, dict) and "ndarray" in dct:
        return np.array(dct["ndarray"])
    return dct


class NPGenerators(enum.Enum):
    herwig = "herwig"
    sherpa = "sherpa"


class NPCalc(law.ExternalTask):
    generator = luigi.EnumParameter(
        enum=NPGenerators,
        description="Generator used for the calculation of NP corrections",
    )
    generator_settings = luigi.Parameter(
        description="Setting String used for the NP calculation. E.g. 'LHC-LO-ZplusJet', 'LHC-NLO-ZplusJet' or 'LHC-ZJetMerging'"
    )
    jet_radius = luigi.Parameter(default="4")

    def output(self):
        targets = {}
        for ybys in ybys_bins:
            targets[ybys] = law.LocalFileTarget(
                "/work/cverstege/QCD_shared/store/PlotNPCorr/"
                + f"{self.generator.value}/{self.generator_settings}/"
                + f"withNP-NPoff-Ratio-Plots/m-ZPt-um-RAW/ZplusJet_3_AK{self.jet_radius}{ybys}.json",
            )
        return law.SiblingFileCollection(targets)


class NPCorrections(law.Task):
    jet_radius = luigi.Parameter(default="4")

    def requires(self):
        req = {}
        for generator in NPGenerators:
            for setting in ("LHC-LO-ZplusJet", "LHC-NLO-ZplusJet", "LHC-ZJetMerging"):
                if setting == "LHC-ZJetMerging" and generator == NPGenerators.herwig:
                    continue
                req[f"{generator.value}_{setting}"] = NPCalc.req(
                    self, generator=generator, generator_settings=setting
                )
        return req

    def output(self):
        ret = {}
        plot_targets = {}
        for ybys in ybys_bins:
            plot_targets[ybys] = law.LocalFileTarget(
                f"{self.__class__.__name__}/" + f"AK{self.jet_radius}_{ybys}.pdf",
                fs="local_plots",
            )
        ret["plots"] = law.SiblingFileCollection(plot_targets)
        ret["root"] = law.LocalFileTarget(
            f"NPCorrections/AK{self.jet_radius}.root",
            fs="local_store",
        )
        return ret

    @safe_output
    def run(self):
        hep.style.use("CMS")
        self.output()["root"].parent.touch()
        out_file = ROOT.TFile.Open(self.output()["root"].abspath, "RECREATE")
        for ybys, _d in ybys_bins.items():
            out_file.mkdir(ybys)
            out_file.cd(ybys)
            correction_dicts = {}
            for setting, target in self.input().items():
                with open(target[ybys].abspath) as fin:
                    data = json.load(fin, object_hook=json_numpy_obj_hook)
                    correction_dicts[setting] = data
            # only derive NP corrections from highest available perturbative order
            corrections = np.stack(
                [
                    correction_dicts[setting]["ys"]
                    for setting in correction_dicts
                    if setting
                    in (
                        "sherpa_LHC-ZJetMerging",
                        "sherpa_LHC-NLO-ZplusJet",
                        "herwig_LHC-NLO-ZplusJet",
                    )
                ],
            )
            corrections_all = np.stack(
                [correction_dicts[setting]["ys"] for setting in correction_dicts],
            )
            uncertainties = np.nan_to_num(
                np.stack(
                    [
                        correction_dicts[setting]["yerrs"]
                        for setting in correction_dicts
                        if setting
                        in (
                            "sherpa_LHC-ZJetMerging",
                            "sherpa_LHC-NLO-ZplusJet",
                            "herwig_LHC-NLO-ZplusJet",
                        )
                    ],
                )
            )
            uncertainties_all = np.nan_to_num(
                np.stack(
                    [
                        correction_dicts[setting]["yerrs"]
                        for setting in correction_dicts
                    ],
                )
            )

            up = np.max(corrections + uncertainties, axis=0)
            down = np.min(corrections - uncertainties, axis=0)
            val = (up + down) / 2
            up_all = np.max(corrections_all + uncertainties_all, axis=0)
            down_all = np.min(corrections_all - uncertainties_all, axis=0)
            val_all = (up_all + down_all) / 2

            # plot the corrections
            fig, ax = plt.subplots()
            hep.label.exp_label(data=True, ax=ax)
            hep.histplot(
                val,
                bins=_d["zpt_binning"],
                histtype="band",
                yerr=[val - down, up - val],
                label=r"NP correction $\pm$ uncertainty",
                edges=False,
                zorder=-2,
            )
            hep.histplot(
                val,
                bins=_d["zpt_binning"],
                histtype="step",
                label=r"NP correction $\pm$ uncertainty",
                edges=False,
                color="black",
                linewidth=4,
                zorder=-1,
            )
            hep.histplot(
                val_all,
                bins=_d["zpt_binning"],
                histtype="band",
                yerr=[val_all - down_all, up_all - val_all],
                label=r"NP correction (alt) $\pm$ uncertainty",
                edges=False,
                hatch="\\\\",
                zorder=-12,
            )
            hep.histplot(
                val_all,
                bins=_d["zpt_binning"],
                histtype="step",
                label=r"NP correction (alt) $\pm$ uncertainty",
                edges=False,
                color="grey",
                linewidth=3,
                zorder=-11,
            )
            for setting, color in zip(self.input().keys(), colormap_5_no_grey):
                label = ""
                if "sherpa" in setting:
                    label += "Sherpa"
                elif "herwig" in setting:
                    label += "Herwig 7"
                if "LHC-ZJetMerging" in setting:
                    label += r" NLO Merged ($\leq 3$ Jets)"
                elif "LHC-LO-ZplusJet" in setting:
                    label += " LO"
                elif "LHC-NLO-ZplusJet" in setting:
                    label += " NLO"
                hep.histplot(
                    correction_dicts[setting]["ys"],
                    bins=_d["zpt_binning"],
                    histtype="errorbar",
                    yerr=correction_dicts[setting]["yerrs"],
                    edges=False,
                    linewidth=2,
                    label=label,
                    markersize=16,
                    elinewidth=3,
                    capsize=4,
                    color=color,
                )
            ax.set_ylabel("NP correction")
            ax.set_xlabel(QUANTITIES["zpt"]["label"])
            ax.set_xscale("log")
            ax.set_xlim(QUANTITIES["zpt"]["range"])
            ax.set_ylim(0.9, 1.5)
            hs, ls = ax.get_legend_handles_labels()
            hs, ls = merge_legend_handles_labels(hs, ls)

            ax.legend(hs, ls, loc="upper right")
            hep.yscale_legend(ax)
            ax.text(
                0.03,
                0.95,
                "${} \\leq y_b < {}$\n".format(*_d["yb_range"])
                + "${} \\leq y^* < {}$".format(*_d["ys_range"]),
                fontsize="small",
                verticalalignment="top",
                fontproperties="Tex Gyre Heros",
                transform=ax.transAxes,
            )
            set_ticklabels(ax)
            fig.tight_layout()
            self.output()["plots"].dir.touch()
            MatplotlibFormatter.dump(self.output()["plots"][ybys], fig)

            # write the corrections to a root file
            assert len(_d["zpt_binning"]) == len(val) + 1
            root_hist = ROOT.TH1D(
                "h_genzpt_npcorr",
                f"{ybys}_genzpt_npcorr",
                len(val),
                _d["zpt_binning"],
            )
            root_hist_up = ROOT.TH1D(
                "h_genzpt_npcorr_up",
                f"{ybys}_genzpt_npcorr_up",
                len(val),
                _d["zpt_binning"],
            )
            root_hist_down = ROOT.TH1D(
                "h_genzpt_npcorr_down",
                f"{ybys}_genzpt_npcorr_down",
                len(val),
                _d["zpt_binning"],
            )
            root_hist_alt = ROOT.TH1D(
                "h_genzpt_npcorr_alt",
                f"{ybys}_genzpt_npcorr_alt",
                len(val),
                _d["zpt_binning"],
            )
            root_hist_up_alt = ROOT.TH1D(
                "h_genzpt_npcorr_alt_up",
                f"{ybys}_genzpt_npcorr_alt_up",
                len(val),
                _d["zpt_binning"],
            )
            root_hist_down_alt = ROOT.TH1D(
                "h_genzpt_npcorr_alt_down",
                f"{ybys}_genzpt_npcorr_alt_down",
                len(val),
                _d["zpt_binning"],
            )
            for i, (v, u, d, va, ua, da) in enumerate(
                zip(val, up, down, val_all, up_all, down_all), start=1
            ):
                root_hist.SetBinContent(i, v)
                root_hist_up.SetBinContent(i, u)
                root_hist_down.SetBinContent(i, d)
                root_hist_alt.SetBinContent(i, va)
                root_hist_up_alt.SetBinContent(i, ua)
                root_hist_down_alt.SetBinContent(i, da)
            root_hist.Write()
            root_hist_up.Write()
            root_hist_down.Write()
            root_hist_alt.Write()
            root_hist_up_alt.Write()
            root_hist_down_alt.Write()
        out_file.Close()
