import os

import matplotlib.pyplot as plt
import numpy as np
import ROOT
import uproot

import law
import luigi
import mplhep as hep
from analysis.framework.quantities import QUANTITIES, ybys_bins
from analysis.tasks.Theory import GridEval, NPCorrections
from analysis.tools.plotting.axes_ticker import set_ticklabels
from analysis.tools.plotting.legend_helpers import merge_legend_handles_labels
from analysis.tools.plotting.styles import theory_styles
from law.contrib.matplotlib import MatplotlibFormatter
from law.decorator import safe_output


class TheoryPrediction(law.Task):
    order = luigi.Parameter(default="NNLO")
    jet_radius = luigi.Parameter(default="4")
    pdf = luigi.Parameter(default="PDF4LHC21_40")

    def requires(self):
        return {"grid": GridEval.req(self), "np": NPCorrections.req(self)}

    def output(self):
        return law.LocalFileTarget(
            f"TheoryPrediction/AK{self.jet_radius}_{self.order}_{self.pdf}.root",
            fs="local_store",
        )

    @safe_output
    def run(self):
        self.output().parent.touch()
        out_file = ROOT.TFile.Open(self.output().abspath, "RECREATE")
        theory_file = uproot.open(self.input()["grid"].abspath)
        np_file = uproot.open(self.input()["np"]["root"].abspath)
        for ybys_bin in ybys_bins:
            out_file.mkdir(ybys_bin)
            out_file.cd(ybys_bin)
            theory_base, bins = theory_file[f"{ybys_bin}/h_genzpt_weight"].to_numpy()
            np_correction, _ = np_file[f"{ybys_bin}/h_genzpt_npcorr"].to_numpy()
            theory_hist = theory_base * np_correction
            out_hists = {"h_genzpt_weight": theory_hist, "h_genzpt_noNP": theory_base}
            for direction in ("up", "down"):
                all_shifts = {}
                noNP_shifts = {}
                np_shift, _ = np_file[
                    f"{ybys_bin}/h_genzpt_npcorr_{direction}"
                ].to_numpy()
                all_shifts[f"h_genzpt_weight_np_{direction}"] = (
                    theory_base * np_shift - theory_hist
                )  # absolute shift, including sign
                for shift in ("scale", "pdf", "stat"):
                    theory_shift, _ = theory_file[
                        f"{ybys_bin}/h_genzpt_weight_{shift}_{direction}"
                    ].to_numpy()
                    noNP_shifts[f"h_genzpt_noNP_{shift}_{direction}"] = theory_shift
                    theory_shift_np = theory_shift * np_correction
                    all_shifts[f"h_genzpt_weight_{shift}_{direction}"] = theory_shift_np
                # caclulate total uncertainty
                total = np.sqrt(
                    np.sum([hist**2 for hist in all_shifts.values()], axis=0)
                )
                total_noNP = np.sqrt(
                    np.sum([hist**2 for hist in noNP_shifts.values()], axis=0)
                )
                if direction == "down":
                    total *= -1
                    total_noNP *= -1
                all_shifts[f"h_genzpt_weight_total_{direction}"] = total
                noNP_shifts[f"h_genzpt_noNP_total_{direction}"] = total_noNP
                out_hists.update(all_shifts)
                out_hists.update(noNP_shifts)
            for name, hist in out_hists.items():
                root_hist = ROOT.TH1D(
                    name,
                    f"{ybys_bin}_{name}",
                    len(bins) - 1,
                    bins,
                )
                for i in range(len(bins) - 1):
                    root_hist.SetBinContent(i + 1, hist[i])
                root_hist.Write()
                del root_hist


class PlotTheoryComparison(law.Task):
    jet_radius = luigi.Parameter(default="4")
    pdf = luigi.Parameter(default="PDF4LHC21_40")

    def requires(self):
        return {
            "LO": TheoryPrediction.req(self, order="LO"),
            "NLO": TheoryPrediction.req(self, order="NLO"),
            "NNLO": TheoryPrediction.req(self, order="NNLO"),
        }

    def output(self):
        plot_targets = {}
        for ybys in ybys_bins:
            plot_targets[ybys] = law.LocalFileTarget(
                f"{self.__class__.__name__}/" + f"AK{self.jet_radius}_{ybys}.pdf",
                fs="local_plots",
            )
        return law.SiblingFileCollection(plot_targets)

    @safe_output
    def run(self):
        hep.style.use("CMS")
        self.output().dir.touch()
        for ybys, _d in ybys_bins.items():
            theories = {}
            theory_uncertainties = {}
            for order in ("LO", "NLO", "NNLO"):
                theory, bins = uproot.open(self.input()[order].abspath)[
                    f"{ybys}/h_genzpt_noNP"
                ].to_numpy()
                theories[order] = theory
                theory_uncertainties[order] = np.zeros(
                    shape=(np.vstack((theory, theory))).shape
                )
                thup, _ = uproot.open(self.input()[order].abspath)[
                    f"{ybys}/h_genzpt_noNP_total_up"
                ].to_numpy()
                thdn, _ = uproot.open(self.input()[order].abspath)[
                    f"{ybys}/h_genzpt_noNP_total_down"
                ].to_numpy()
                theory_uncertainties[order] = np.abs(np.vstack((thdn, thup)))
            # plot the ratios to LO
            fig, ax = plt.subplots()
            hep.label.exp_label(data=True, ax=ax)
            for order in ("LO", "NLO", "NNLO"):
                style = theory_styles[order]
                hep.histplot(
                    theories[order] / theories["LO"],
                    bins=_d["zpt_binning"],
                    histtype="step",
                    label=order,
                    edges=False,
                    linewidth=4,
                    color=style["color"],
                )
                hep.histplot(
                    theories[order] / theories["LO"],
                    bins=_d["zpt_binning"],
                    histtype="band",
                    yerr=theory_uncertainties[order] / theories["LO"],
                    label=order,
                    edges=False,
                    edgecolor=style["color"],
                )

            ax.set_ylabel("Ratio to LO")
            ax.set_xlabel(QUANTITIES["zpt"]["label"])
            ax.set_xscale("log")
            ax.set_xlim(QUANTITIES["zpt"]["range"])
            # ax.set_ylim(0.5, 20)
            # ax.set_yscale("log")
            hs, ls = ax.get_legend_handles_labels()
            hs, ls = merge_legend_handles_labels(hs, ls)

            ax.legend(hs, ls, loc="upper right")
            hep.yscale_legend(ax)
            ax.text(
                0.03,
                0.95,
                "${} \\leq y_b < {}$\n".format(*_d["yb_range"])
                + "${} \\leq y^* < {}$".format(*_d["ys_range"]),
                fontsize="small",
                verticalalignment="top",
                fontproperties="Tex Gyre Heros",
                transform=ax.transAxes,
            )
            set_ticklabels(ax)
            fig.tight_layout()
            self.output().dir.touch()
            MatplotlibFormatter.dump(self.output()[ybys], fig)


class PlotTheoryUncertainties(law.Task):
    order = luigi.Parameter(default="NNLO")
    jet_radius = luigi.Parameter(default="4")
    pdf = luigi.Parameter(default="PDF4LHC21_40")

    def output(self):
        plot_targets = {}
        for ybys in ybys_bins:
            plot_targets[ybys] = law.LocalFileTarget(
                os.path.join(
                    self.__class__.__name__.replace("Plot", ""),
                    f"{self.order}_{self.pdf}",
                    f"AK{self.jet_radius}_{ybys}.pdf",
                ),
                fs="local_plots",
            )
        return law.SiblingFileCollection(plot_targets)

    def requires(self):
        return TheoryPrediction.req(self)

    def run(self):
        hep.style.use("CMS")
        self.output().dir.touch()
        theory_file = uproot.open(self.input().abspath)
        uncs = {
            "total": dict(
                label="Total theory unc.", color="black", linestyle="solid", linewidth=3
            ),
            "stat": dict(label="Stat. unc.", color="#9c9ca1"),
            "scale": dict(label="Scale", color="#f89c20", linestyle="dashdot"),
            "pdf": dict(label=f"{self.pdf}", color="#5790fc", linestyle="dotted"),
            "np": dict(label="NP corr.", color="#964a8b", linestyle="dashed"),
        }
        for ybys, _d in ybys_bins.items():
            fig, ax = plt.subplots()
            theory, bins = theory_file[f"{ybys}/h_genzpt_weight"].to_numpy()
            hep.label.exp_label(data=True, ax=ax)
            for unc, unc_dict in uncs.items():
                up = theory_file[f"{ybys}/h_genzpt_weight_{unc}_up"].to_numpy()[0]
                dn = theory_file[f"{ybys}/h_genzpt_weight_{unc}_down"].to_numpy()[0]
                ax.stairs(
                    up / theory,
                    edges=bins,
                    label=unc_dict["label"],
                    color=unc_dict["color"],
                    linestyle=unc_dict.get("linestyle", "-"),
                    linewidth=unc_dict.get("linewidth", 2),
                    baseline=None,
                )
                ax.stairs(
                    dn / theory,
                    edges=bins,
                    color=unc_dict["color"],
                    linestyle=unc_dict.get("linestyle", "-"),
                    linewidth=unc_dict.get("linewidth", 2),
                    baseline=None,
                )
            ax.set_ylabel("Relative Uncertainty")
            ax.set_xlabel(QUANTITIES["zpt"]["label"])
            ax.set_xscale("log")
            ax.set_xlim(QUANTITIES["zpt"]["range"])
            ax.set_ylim(-0.25, 0.3)
            ax.hlines(0, 25, 1000, color="black", linestyle="dashed")
            ax.text(
                0.03,
                0.95,
                _d["label"],
                fontsize="small",
                verticalalignment="top",
                fontproperties="Tex Gyre Heros",
                transform=ax.transAxes,
            )
            set_ticklabels(ax)
            ax.legend(loc="upper right")
            fig.tight_layout()
            MatplotlibFormatter.dump(self.output()[ybys], fig)
