from matplotlib.patches import Rectangle
from matplotlib.ticker import FixedLocator, NullLocator


def _simpleaxis(ax, invert_x=False, invert_y=True, x_tick_label_rotation=0):
    # ax.set_frame_on(False)
    ax.patch.set_alpha(0.0)
    _vis_x, _invis_x = ("bottom", "top")
    _vis_y, _invis_y = ("left", "right")
    if invert_x:
        _vis_y, _invis_y = _invis_y, _vis_y
    if invert_y:
        _vis_x, _invis_x = _invis_x, _vis_x

    # hide/show the right spines/labels/ticks
    for _axis, _s in zip((ax.xaxis, ax.yaxis), (_vis_x, _vis_y)):
        _axis.set_ticks_position(_s)
        _axis.set_label_position(_s)
        ax.spines[_s].set_visible(True)

    if x_tick_label_rotation:
        _ha = (
            "center"
            if x_tick_label_rotation == 0
            else ("left" if invert_y == (x_tick_label_rotation > 0) else "right")
        )
        ax.xaxis.set_ticklabels(
            ax.xaxis.get_ticklabels(), rotation=x_tick_label_rotation, ha=_ha
        )

    for _s in _invis_x, _invis_y:
        ax.spines[_s].set_visible(False)

    # # hide/show the right ticks (needed?)
    # ax.get_xaxis().tick_top() if invert_y else ax.get_xaxis().tick_bottom()
    # ax.get_yaxis().tick_right() if invert_x else ax.get_yaxis().tick_left()


def custom_ybys_legend(
    ax,
    ybys_dicts,
    size_inches=2.0,
    borderpad=5,
    loc="upper_right",
    use_inset=True,
    vertical_axis="ys",
    bin_label_key=None,
    x_tick_label_rotation=0,
    shaded_bin_backgrounds=True,
    fontsize="x-small",
):
    _LOCS = {
        "upper right": dict(id_=1, invert_x=True, invert_y=True),
        "upper left": dict(id_=2, invert_x=False, invert_y=True),
        "lower right": dict(id_=4, invert_x=True, invert_y=False),
        "lower left": dict(id_=3, invert_x=False, invert_y=False),
        "center right": dict(id_=7, invert_x=False, invert_y=False),
    }
    _loc = _LOCS[loc]

    if vertical_axis == "ys":
        _qx, _qy = "yb", "ys"
    elif vertical_axis == "yb":
        _qx, _qy = "ys", "yb"
    else:
        raise ValueError("Vertical axis spec '{}' is not one of: 'ys', 'yb'")

    if use_inset:
        from mpl_toolkits.axes_grid1.inset_locator import inset_axes

        ax = inset_axes(
            ax,
            width=size_inches,
            height=size_inches,
            loc=_loc["id_"],
            borderpad=borderpad,
        )

    _y_sign = -1.0 if _loc["invert_y"] else 1.0
    _x_sign = -1.0 if _loc["invert_x"] else 1.0

    ax.axis("equal")
    for _s in ybys_dicts:
        _m_fill_style = _s.get("marker_style", "full")
        _m_fill_color_alt = "none"
        if _m_fill_style == "empty":
            _m_fill_style = "none"
            _m_fill_color_alt = "yellow"
        elif _m_fill_style == "white":
            _m_fill_style = "none"
            _m_fill_color_alt = "w"

        _plot_marker_kwargs = dict(
            markerfacecoloralt=_m_fill_color_alt,
            markeredgewidth=1,
            fillstyle=_m_fill_style,
        )

        if bin_label_key is not None:
            ax.annotate(
                text=_s.get(bin_label_key, "?"),
                xy=(
                    _x_sign * _s["{}_mean".format(_qx)] - 0.0,
                    _y_sign * _s["{}_mean".format(_qy)] - 0.13,
                ),
                ha="center",
                va="center",
                fontsize=fontsize,
            )
            ax.plot(
                [_x_sign * _s["{}_mean".format(_qx)] + 0.0],
                [_y_sign * _s["{}_mean".format(_qy)] + 0.12],
                color=_s.get("color", None),
                marker=_s.get("marker", "s"),
                # markersize=_s.get('markersize', 5),  # * float(5)/7,
                **_plot_marker_kwargs,
            )
        else:
            ax.plot(
                [_x_sign * _s["{}_mean".format(_qx)]],
                [_y_sign * _s["{}_mean".format(_qy)]],
                color=_s.get("color", None),
                marker=_s.get("marker", "s"),
                # markersize=_s.get('markersize', 5),
                **_plot_marker_kwargs,
            )

        if shaded_bin_backgrounds:
            _pad = 0.03
            _w = _x_sign * (
                _s["{}_range".format(_qx)][1] - _s["{}_range".format(_qx)][0] - 2 * _pad
            )
            _h = _y_sign * (
                _s["{}_range".format(_qy)][1] - _s["{}_range".format(_qy)][0] - 2 * _pad
            )
            r = Rectangle(
                (
                    _x_sign * (_s["{}_range".format(_qx)][0] + _pad),
                    _y_sign * (_s["{}_range".format(_qy)][0] + _pad),
                ),
                width=_w,
                height=_h,
                color=_s["color"],
                alpha=0.2,
            )
            ax.add_patch(r)

    if _loc["invert_x"]:
        ax.set_xlim(-2.6, 0.1)
    else:
        ax.set_xlim(-0.1, 2.6)

    if _loc["invert_y"]:
        ax.set_ylim(-2.6, 0.1)
    else:
        ax.set_ylim(-0.1, 2.6)

    _labels = {
        "yb": r"$\mathit{y}_\mathdefault{b}$",
        "ys": r"$\mathit{y}\mathdefault{*}$",
    }

    # determine need to shift x axis label so it doesn't overlap with rotated tick labels
    # _ha = (
    #     "center"
    #     if x_tick_label_rotation == 0
    #     else ("left" if _loc["invert_y"] == (x_tick_label_rotation > 0) else "right")
    # )
    # _x_axis_label_offset = (
    #     ((_ha == "left") == (not _loc["invert_x"]))
    #     * 0.15
    #     * np.sin(x_tick_label_rotation / 180.0 * np.pi)
    # )

    ax.xaxis.set_major_locator(FixedLocator([_x_sign * i / 2.0 for i in range(6)]))
    ax.yaxis.set_major_locator(FixedLocator([_y_sign * i / 2.0 for i in range(6)]))
    ax.xaxis.set_minor_locator(NullLocator())
    ax.yaxis.set_minor_locator(NullLocator())

    ax.set_xticklabels([str(i / 2.0) for i in range(6)], size=fontsize)
    ax.set_yticklabels([str(i / 2.0) for i in range(6)], size=fontsize)

    _simpleaxis(
        ax,
        invert_x=_loc["invert_x"],
        invert_y=_loc["invert_y"],
        x_tick_label_rotation=x_tick_label_rotation,
    )

    ax.set_xlabel(
        _labels[_qx],
        x=-0.005 if _loc["invert_x"] else 1.005,  # + _x_axis_label_offset,
        ha="left" if _loc["invert_x"] else "right",
        va="top" if _loc["invert_y"] else "bottom",
        size=fontsize,
    )
    ax.set_ylabel(
        _labels[_qy],
        y=-0.005 if _loc["invert_y"] else 1.005,
        ha="right" if _loc["invert_y"] else "left",
        va="bottom" if _loc["invert_x"] else "top",
        # rotation=0, rotation_mode='default',
        size=fontsize,
    )
